@@ -4,8 +4,7 @@ use std::ops::Deref;
 use std::fmt;
 
 use crate::compiler::tokens::Span;
-use crate::key::Key;
-use crate::value::{value_map_with_capacity, MapType, Value, ValueRepr};
+use crate::value::{value_map_with_capacity, Value};
 
 /// Container for nodes with location info.
 ///
@@ -81,6 +80,10 @@ pub enum Stmt<'a> {
     Macro(Spanned<Macro<'a>>),
     #[cfg(feature = "macros")]
     CallBlock(Spanned<CallBlock<'a>>),
+    #[cfg(feature = "loop_controls")]
+    Continue(Spanned<Continue>),
+    #[cfg(feature = "loop_controls")]
+    Break(Spanned<Break>),
     Do(Spanned<Do<'a>>),
 }
 
@@ -112,6 +115,10 @@ impl<'a> fmt::Debug for Stmt<'a> {
             Stmt::Macro(s) => fmt::Debug::fmt(s, f),
             #[cfg(feature = "macros")]
             Stmt::CallBlock(s) => fmt::Debug::fmt(s, f),
+            #[cfg(feature = "loop_controls")]
+            Stmt::Continue(s) => fmt::Debug::fmt(s, f),
+            #[cfg(feature = "loop_controls")]
+            Stmt::Break(s) => fmt::Debug::fmt(s, f),
             Stmt::Do(s) => fmt::Debug::fmt(s, f),
         }
     }
@@ -138,7 +145,6 @@ pub enum Expr<'a> {
     Call(Spanned<Call<'a>>),
     List(Spanned<List<'a>>),
     Map(Spanned<Map<'a>>),
-    Kwargs(Spanned<Kwargs<'a>>),
 }
 
 #[cfg(feature = "internal_debug")]
@@ -158,7 +164,6 @@ impl<'a> fmt::Debug for Expr<'a> {
             Expr::Call(s) => fmt::Debug::fmt(s, f),
             Expr::List(s) => fmt::Debug::fmt(s, f),
             Expr::Map(s) => fmt::Debug::fmt(s, f),
-            Expr::Kwargs(s) => fmt::Debug::fmt(s, f),
         }
     }
 }
@@ -179,7 +184,6 @@ impl<'a> Expr<'a> {
             Expr::Map(_) => "map literal",
             Expr::Test(_) => "test expression",
             Expr::Filter(_) => "filter expression",
-            Expr::Kwargs(_) => "keyword arguments",
         }
     }
 }
@@ -299,6 +303,18 @@ pub struct CallBlock<'a> {
     pub macro_decl: Spanned<Macro<'a>>,
 }
 
+/// Continue
+#[cfg_attr(feature = "internal_debug", derive(Debug))]
+#[cfg(feature = "loop_controls")]
+#[cfg_attr(feature = "unstable_machinery_serde", derive(serde::Serialize))]
+pub struct Continue;
+
+/// Break
+#[cfg_attr(feature = "internal_debug", derive(Debug))]
+#[cfg(feature = "loop_controls")]
+#[cfg_attr(feature = "unstable_machinery_serde", derive(serde::Serialize))]
+pub struct Break;
+
 /// A call block
 #[cfg_attr(feature = "internal_debug", derive(Debug))]
 #[cfg_attr(feature = "unstable_machinery_serde", derive(serde::Serialize))]
@@ -425,7 +441,7 @@ pub struct IfExpr<'a> {
 pub struct Filter<'a> {
     pub name: &'a str,
     pub expr: Option<Expr<'a>>,
-    pub args: Vec<Expr<'a>>,
+    pub args: Vec<CallArg<'a>>,
 }
 
 /// A test expression.
@@ -434,7 +450,7 @@ pub struct Filter<'a> {
 pub struct Test<'a> {
     pub name: &'a str,
     pub expr: Expr<'a>,
-    pub args: Vec<Expr<'a>>,
+    pub args: Vec<CallArg<'a>>,
 }
 
 /// An attribute lookup expression.
@@ -458,7 +474,17 @@ pub struct GetItem<'a> {
 #[cfg_attr(feature = "unstable_machinery_serde", derive(serde::Serialize))]
 pub struct Call<'a> {
     pub expr: Expr<'a>,
-    pub args: Vec<Expr<'a>>,
+    pub args: Vec<CallArg<'a>>,
+}
+
+/// A call argument helper
+#[cfg_attr(feature = "internal_debug", derive(Debug))]
+#[cfg_attr(feature = "unstable_machinery_serde", derive(serde::Serialize))]
+pub enum CallArg<'a> {
+    Pos(Expr<'a>),
+    Kwarg(&'a str, Expr<'a>),
+    PosSplat(Expr<'a>),
+    KwargSplat(Expr<'a>),
 }
 
 /// Creates a list of values.
@@ -480,31 +506,7 @@ impl<'a> List<'a> {
             _ => None,
         });
 
-        Some(sequence.collect())
-    }
-}
-
-/// Creates a map of kwargs
-#[cfg_attr(feature = "internal_debug", derive(Debug))]
-#[cfg_attr(feature = "unstable_machinery_serde", derive(serde::Serialize))]
-pub struct Kwargs<'a> {
-    pub pairs: Vec<(&'a str, Expr<'a>)>,
-}
-
-impl<'a> Kwargs<'a> {
-    pub fn as_const(&self) -> Option<Value> {
-        if !self.pairs.iter().all(|x| matches!(x.1, Expr::Const(_))) {
-            return None;
-        }
-
-        let mut rv = value_map_with_capacity(self.pairs.len());
-        for (key, value) in &self.pairs {
-            if let Expr::Const(value) = value {
-                rv.insert(Key::make_string_key(key), value.value.clone());
-            }
-        }
-
-        Some(Value(ValueRepr::Map(rv.into(), MapType::Kwargs)))
+        Some(Value::from(sequence.collect::<Vec<_>>()))
     }
 }
 
@@ -527,17 +529,11 @@ impl<'a> Map<'a> {
         let mut rv = value_map_with_capacity(self.keys.len());
         for (key, value) in self.keys.iter().zip(self.values.iter()) {
             if let (Expr::Const(maybe_key), Expr::Const(value)) = (key, value) {
-                rv.insert(
-                    match maybe_key.value.clone().try_into_key() {
-                        Ok(key) => key,
-                        Err(_) => return None,
-                    },
-                    value.value.clone(),
-                );
+                rv.insert(maybe_key.value.clone(), value.value.clone());
             }
         }
 
-        Some(Value(ValueRepr::Map(rv.into(), MapType::Normal)))
+        Some(Value::from_object(rv))
     }
 }
 
