@@ -6,7 +6,8 @@ use crate::compiler::instructions::{
 };
 use crate::compiler::tokens::Span;
 use crate::output::CaptureMode;
-use crate::value::Value;
+use crate::value::ops::neg;
+use crate::value::{Kwargs, Value, ValueMap};
 
 #[cfg(test)]
 use similar_asserts::assert_eq;
@@ -33,9 +34,16 @@ fn get_local_id<'source>(ids: &mut BTreeMap<&'source str, LocalId>, name: &'sour
 /// Represents an open block of code that does not yet have updated
 /// jump targets.
 enum PendingBlock {
-    Branch(usize),
-    Loop(usize),
-    ScBool(Vec<usize>),
+    Branch {
+        jump_instr: usize,
+    },
+    Loop {
+        iter_instr: usize,
+        jump_instrs: Vec<usize>,
+    },
+    ScBool {
+        jump_instrs: Vec<usize>,
+    },
 }
 
 /// Provides a convenient interface to creating instructions for the VM.
@@ -120,7 +128,7 @@ impl<'source> CodeGenerator<'source> {
     fn finish_subgenerator(&mut self, sub: CodeGenerator<'source>) -> Instructions<'source> {
         self.current_line = sub.current_line;
         let (instructions, blocks) = sub.finish();
-        self.blocks.extend(blocks.into_iter());
+        self.blocks.extend(blocks);
         instructions
     }
 
@@ -134,43 +142,51 @@ impl<'source> CodeGenerator<'source> {
             flags |= LOOP_FLAG_RECURSIVE;
         }
         self.add(Instruction::PushLoop(flags));
-        let iter_instr = self.add(Instruction::Iterate(!0));
-        self.pending_block.push(PendingBlock::Loop(iter_instr));
+        let instr = self.add(Instruction::Iterate(!0));
+        self.pending_block.push(PendingBlock::Loop {
+            iter_instr: instr,
+            jump_instrs: Vec::new(),
+        });
     }
 
     /// Ends the open for loop
     pub fn end_for_loop(&mut self, push_did_not_iterate: bool) {
-        match self.pending_block.pop() {
-            Some(PendingBlock::Loop(iter_instr)) => {
-                self.add(Instruction::Jump(iter_instr));
-                let loop_end = self.next_instruction();
-                if push_did_not_iterate {
-                    self.add(Instruction::PushDidNotIterate);
-                };
-                self.add(Instruction::PopFrame);
-                if let Some(Instruction::Iterate(ref mut jump_target)) =
-                    self.instructions.get_mut(iter_instr)
-                {
-                    *jump_target = loop_end;
-                } else {
-                    unreachable!();
+        if let Some(PendingBlock::Loop {
+            iter_instr,
+            jump_instrs,
+        }) = self.pending_block.pop()
+        {
+            self.add(Instruction::Jump(iter_instr));
+            let loop_end = self.next_instruction();
+            if push_did_not_iterate {
+                self.add(Instruction::PushDidNotIterate);
+            };
+            self.add(Instruction::PopFrame);
+            for instr in jump_instrs.into_iter().chain(Some(iter_instr)) {
+                match self.instructions.get_mut(instr) {
+                    Some(Instruction::Iterate(ref mut jump_target))
+                    | Some(Instruction::Jump(ref mut jump_target)) => {
+                        *jump_target = loop_end;
+                    }
+                    _ => unreachable!(),
                 }
             }
-            _ => unreachable!(),
+        } else {
+            unreachable!()
         }
     }
 
     /// Begins an if conditional
     pub fn start_if(&mut self) {
         let jump_instr = self.add(Instruction::JumpIfFalse(!0));
-        self.pending_block.push(PendingBlock::Branch(jump_instr));
+        self.pending_block.push(PendingBlock::Branch { jump_instr });
     }
 
     /// Begins an else conditional
     pub fn start_else(&mut self) {
         let jump_instr = self.add(Instruction::Jump(!0));
         self.end_condition(jump_instr + 1);
-        self.pending_block.push(PendingBlock::Branch(jump_instr));
+        self.pending_block.push(PendingBlock::Branch { jump_instr });
     }
 
     /// Closes the current if block.
@@ -178,15 +194,20 @@ impl<'source> CodeGenerator<'source> {
         self.end_condition(self.next_instruction());
     }
 
-    /// Starts a short cirquited bool block.
+    /// Starts a short-circuited bool block.
     pub fn start_sc_bool(&mut self) {
-        self.pending_block.push(PendingBlock::ScBool(vec![]));
+        self.pending_block.push(PendingBlock::ScBool {
+            jump_instrs: Vec::new(),
+        });
     }
 
-    /// Emits a short circuited bool operator.
+    /// Emits a short-circuited bool operator.
     pub fn sc_bool(&mut self, and: bool) {
-        if let Some(PendingBlock::ScBool(ref mut instructions)) = self.pending_block.last_mut() {
-            instructions.push(self.instructions.add(if and {
+        if let Some(PendingBlock::ScBool {
+            ref mut jump_instrs,
+        }) = self.pending_block.last_mut()
+        {
+            jump_instrs.push(self.instructions.add(if and {
                 Instruction::JumpIfFalseOrPop(!0)
             } else {
                 Instruction::JumpIfTrueOrPop(!0)
@@ -196,11 +217,11 @@ impl<'source> CodeGenerator<'source> {
         }
     }
 
-    /// Ends a short circuited bool block.
+    /// Ends a short-circuited bool block.
     pub fn end_sc_bool(&mut self) {
         let end = self.next_instruction();
-        if let Some(PendingBlock::ScBool(instructions)) = self.pending_block.pop() {
-            for instr in instructions {
+        if let Some(PendingBlock::ScBool { jump_instrs }) = self.pending_block.pop() {
+            for instr in jump_instrs {
                 match self.instructions.get_mut(instr) {
                     Some(Instruction::JumpIfFalseOrPop(ref mut target))
                     | Some(Instruction::JumpIfTrueOrPop(ref mut target)) => {
@@ -212,15 +233,17 @@ impl<'source> CodeGenerator<'source> {
         }
     }
 
-    fn end_condition(&mut self, jump_instr: usize) {
+    fn end_condition(&mut self, new_jump_instr: usize) {
         match self.pending_block.pop() {
-            Some(PendingBlock::Branch(instr)) => match self.instructions.get_mut(instr) {
-                Some(Instruction::JumpIfFalse(ref mut target))
-                | Some(Instruction::Jump(ref mut target)) => {
-                    *target = jump_instr;
+            Some(PendingBlock::Branch { jump_instr }) => {
+                match self.instructions.get_mut(jump_instr) {
+                    Some(Instruction::JumpIfFalse(ref mut target))
+                    | Some(Instruction::Jump(ref mut target)) => {
+                        *target = new_jump_instr;
+                    }
+                    _ => {}
                 }
-                _ => {}
-            },
+            }
             _ => unreachable!(),
         }
     }
@@ -346,6 +369,31 @@ impl<'source> CodeGenerator<'source> {
             ast::Stmt::CallBlock(call_block) => {
                 self.compile_call_block(call_block);
             }
+            #[cfg(feature = "loop_controls")]
+            ast::Stmt::Continue(cont) => {
+                self.set_line_from_span(cont.span());
+                for pending_block in self.pending_block.iter().rev() {
+                    if let PendingBlock::Loop { iter_instr, .. } = pending_block {
+                        self.add(Instruction::Jump(*iter_instr));
+                        break;
+                    }
+                }
+            }
+            #[cfg(feature = "loop_controls")]
+            ast::Stmt::Break(brk) => {
+                self.set_line_from_span(brk.span());
+                let instr = self.add(Instruction::Jump(0));
+                for pending_block in self.pending_block.iter_mut().rev() {
+                    if let PendingBlock::Loop {
+                        ref mut jump_instrs,
+                        ..
+                    } = pending_block
+                    {
+                        jump_instrs.push(instr);
+                        break;
+                    }
+                }
+            }
             ast::Stmt::Do(do_tag) => {
                 self.compile_do(do_tag);
             }
@@ -367,7 +415,6 @@ impl<'source> CodeGenerator<'source> {
     #[cfg(feature = "macros")]
     fn compile_macro_expression(&mut self, macro_decl: &ast::Spanned<ast::Macro<'source>>) {
         use crate::compiler::instructions::MACRO_CALLER;
-        use crate::value::ValueRepr;
         self.set_line_from_span(macro_decl.span());
         let instr = self.add(Instruction::Jump(!0));
         let mut defaults_iter = macro_decl.defaults.iter().rev();
@@ -393,7 +440,7 @@ impl<'source> CodeGenerator<'source> {
             self.add(Instruction::Enclose(name));
         }
         self.add(Instruction::GetClosure);
-        self.add(Instruction::LoadConst(Value::from(ValueRepr::Seq(
+        self.add(Instruction::LoadConst(Value::from_object(
             macro_decl
                 .args
                 .iter()
@@ -401,9 +448,8 @@ impl<'source> CodeGenerator<'source> {
                     ast::Expr::Var(var) => Value::from(var.id),
                     _ => unreachable!(),
                 })
-                .collect::<Vec<_>>()
-                .into(),
-        ))));
+                .collect::<Vec<Value>>(),
+        )));
         let mut flags = 0;
         if caller_reference {
             flags |= MACRO_CALLER;
@@ -457,7 +503,7 @@ impl<'source> CodeGenerator<'source> {
                         self.add_with_span(Instruction::FastSuper, call.span());
                         return;
                     } else if name == "loop" && call.args.len() == 1 {
-                        self.compile_expr(&call.args[0]);
+                        self.compile_call_args(std::slice::from_ref(&call.args[0]), 0, None);
                         self.add(Instruction::FastRecurse);
                         return;
                     }
@@ -476,25 +522,31 @@ impl<'source> CodeGenerator<'source> {
 
     fn compile_for_loop(&mut self, for_loop: &ast::Spanned<ast::ForLoop<'source>>) {
         self.set_line_from_span(for_loop.span());
+
+        // filter expressions work like a nested for loop without
+        // the special loop variable. in one loop, the condition is checked and
+        // passing items accumulated into a list. in the second, that list is
+        // iterated over normally
         if let Some(ref filter_expr) = for_loop.filter_expr {
-            // filter expressions work like a nested for loop without
-            // the special loop variable that append into a new list
-            // just outside of the loop.
-            self.add(Instruction::BuildList(0));
+            self.add(Instruction::LoadConst(Value::from(0usize)));
             self.compile_expr(&for_loop.iter);
             self.start_for_loop(false, false);
             self.add(Instruction::DupTop);
             self.compile_assignment(&for_loop.target);
             self.compile_expr(filter_expr);
             self.start_if();
-            self.add(Instruction::ListAppend);
+            self.add(Instruction::Swap);
+            self.add(Instruction::LoadConst(Value::from(1usize)));
+            self.add(Instruction::Add);
             self.start_else();
             self.add(Instruction::DiscardTop);
             self.end_if();
             self.end_for_loop(false);
+            self.add(Instruction::BuildList(None));
         } else {
             self.compile_expr(&for_loop.iter);
         }
+
         self.start_for_loop(true, for_loop.recursive);
         self.compile_assignment(&for_loop.target);
         for node in &for_loop.body {
@@ -524,6 +576,11 @@ impl<'source> CodeGenerator<'source> {
                 }
                 self.pop_span();
             }
+            ast::Expr::GetAttr(attr) => {
+                self.push_span(attr.span());
+                self.compile_expr(&attr.expr);
+                self.add(Instruction::SetAttr(attr.name));
+            }
             _ => unreachable!(),
         }
     }
@@ -562,11 +619,25 @@ impl<'source> CodeGenerator<'source> {
             }
             ast::Expr::UnaryOp(c) => {
                 self.set_line_from_span(c.span());
-                self.compile_expr(&c.expr);
                 match c.op {
-                    ast::UnaryOpKind::Not => self.add(Instruction::Not),
-                    ast::UnaryOpKind::Neg => self.add_with_span(Instruction::Neg, c.span()),
-                };
+                    ast::UnaryOpKind::Not => {
+                        self.compile_expr(&c.expr);
+                        self.add(Instruction::Not);
+                    }
+                    ast::UnaryOpKind::Neg => {
+                        // common case: negative numbers.  In that case we
+                        // directly negate them if this is possible without
+                        // an error.
+                        if let ast::Expr::Const(ref c) = c.expr {
+                            if let Ok(negated) = neg(&c.value) {
+                                self.add(Instruction::LoadConst(negated));
+                                return;
+                            }
+                        }
+                        self.compile_expr(&c.expr);
+                        self.add_with_span(Instruction::Neg, c.span());
+                    }
+                }
             }
             ast::Expr::BinOp(c) => {
                 self.compile_bin_op(c);
@@ -589,21 +660,17 @@ impl<'source> CodeGenerator<'source> {
                 if let Some(ref expr) = f.expr {
                     self.compile_expr(expr);
                 }
-                for arg in &f.args {
-                    self.compile_expr(arg);
-                }
+                let arg_count = self.compile_call_args(&f.args, 1, None);
                 let local_id = get_local_id(&mut self.filter_local_ids, f.name);
-                self.add(Instruction::ApplyFilter(f.name, f.args.len() + 1, local_id));
+                self.add(Instruction::ApplyFilter(f.name, arg_count, local_id));
                 self.pop_span();
             }
             ast::Expr::Test(f) => {
                 self.push_span(f.span());
                 self.compile_expr(&f.expr);
-                for arg in &f.args {
-                    self.compile_expr(arg);
-                }
+                let arg_count = self.compile_call_args(&f.args, 1, None);
                 let local_id = get_local_id(&mut self.test_local_ids, f.name);
-                self.add(Instruction::PerformTest(f.name, f.args.len() + 1, local_id));
+                self.add(Instruction::PerformTest(f.name, arg_count, local_id));
                 self.pop_span();
             }
             ast::Expr::GetAttr(g) => {
@@ -630,7 +697,7 @@ impl<'source> CodeGenerator<'source> {
                     for item in &l.items {
                         self.compile_expr(item);
                     }
-                    self.add(Instruction::BuildList(l.items.len()));
+                    self.add(Instruction::BuildList(Some(l.items.len())));
                 }
             }
             ast::Expr::Map(m) => {
@@ -646,18 +713,6 @@ impl<'source> CodeGenerator<'source> {
                     self.add(Instruction::BuildMap(m.keys.len()));
                 }
             }
-            ast::Expr::Kwargs(m) => {
-                if let Some(val) = m.as_const() {
-                    self.add(Instruction::LoadConst(val));
-                } else {
-                    self.set_line_from_span(m.span());
-                    for (key, value) in &m.pairs {
-                        self.add(Instruction::LoadConst(Value::from(*key)));
-                        self.compile_expr(value);
-                    }
-                    self.add(Instruction::BuildKwargs(m.pairs.len()));
-                }
-            }
         }
     }
 
@@ -669,7 +724,7 @@ impl<'source> CodeGenerator<'source> {
         self.push_span(c.span());
         match c.identify_call() {
             ast::CallType::Function(name) => {
-                let arg_count = self.compile_call_args(&c.args, caller);
+                let arg_count = self.compile_call_args(&c.args, 0, caller);
                 self.add(Instruction::CallFunction(name, arg_count));
             }
             #[cfg(feature = "multi_template")]
@@ -680,13 +735,13 @@ impl<'source> CodeGenerator<'source> {
             }
             ast::CallType::Method(expr, name) => {
                 self.compile_expr(expr);
-                let arg_count = self.compile_call_args(&c.args, caller);
-                self.add(Instruction::CallMethod(name, arg_count + 1));
+                let arg_count = self.compile_call_args(&c.args, 1, caller);
+                self.add(Instruction::CallMethod(name, arg_count));
             }
             ast::CallType::Object(expr) => {
                 self.compile_expr(expr);
-                let arg_count = self.compile_call_args(&c.args, caller);
-                self.add(Instruction::CallObject(arg_count + 1));
+                let arg_count = self.compile_call_args(&c.args, 1, caller);
+                self.add(Instruction::CallObject(arg_count));
             }
         };
         self.pop_span();
@@ -694,57 +749,112 @@ impl<'source> CodeGenerator<'source> {
 
     fn compile_call_args(
         &mut self,
-        args: &[ast::Expr<'source>],
+        args: &[ast::CallArg<'source>],
+        extra_args: usize,
         caller: Option<&Caller<'source>>,
-    ) -> usize {
-        match caller {
-            // we can conditionally compile the caller part here since this will
-            // nicely call through for non macro builds
-            #[cfg(feature = "macros")]
-            Some(caller) => self.compile_call_args_with_caller(args, caller),
-            _ => {
-                for arg in args {
-                    self.compile_expr(arg);
-                }
-                args.len()
-            }
-        }
-    }
+    ) -> Option<u16> {
+        let mut pending_args = extra_args;
+        let mut num_args_batches = 0;
+        let mut has_kwargs = caller.is_some();
+        let mut static_kwargs = caller.is_none();
 
-    #[cfg(feature = "macros")]
-    fn compile_call_args_with_caller(
-        &mut self,
-        args: &[ast::Expr<'source>],
-        caller: &Caller<'source>,
-    ) -> usize {
-        let mut injected_caller = false;
-
-        // try to add the caller to already existing keyword arguments.
         for arg in args {
-            if let ast::Expr::Kwargs(ref m) = arg {
-                self.set_line_from_span(m.span());
-                for (key, value) in &m.pairs {
-                    self.add(Instruction::LoadConst(Value::from(*key)));
-                    self.compile_expr(value);
+            match arg {
+                ast::CallArg::Pos(expr) => {
+                    self.compile_expr(expr);
+                    pending_args += 1;
+                }
+                ast::CallArg::PosSplat(expr) => {
+                    if pending_args > 0 {
+                        self.add(Instruction::BuildList(Some(pending_args)));
+                        pending_args = 0;
+                        num_args_batches += 1;
+                    }
+                    self.compile_expr(expr);
+                    num_args_batches += 1;
+                }
+                ast::CallArg::Kwarg(_, expr) => {
+                    if !matches!(expr, ast::Expr::Const(_)) {
+                        static_kwargs = false;
+                    }
+                    has_kwargs = true;
+                }
+                ast::CallArg::KwargSplat(_) => {
+                    static_kwargs = false;
+                    has_kwargs = true;
                 }
-                self.add(Instruction::LoadConst(Value::from("caller")));
-                self.compile_macro_expression(caller);
-                self.add(Instruction::BuildKwargs(m.pairs.len() + 1));
-                injected_caller = true;
-            } else {
-                self.compile_expr(arg);
             }
         }
 
-        // if there are no keyword args so far, create a new kwargs object
-        // and add caller to that.
-        if !injected_caller {
-            self.add(Instruction::LoadConst(Value::from("caller")));
-            self.compile_macro_expression(caller);
-            self.add(Instruction::BuildKwargs(1));
-            args.len() + 1
+        if has_kwargs {
+            let mut pending_kwargs = 0;
+            let mut num_kwargs_batches = 0;
+            let mut collected_kwargs = ValueMap::new();
+            for arg in args {
+                match arg {
+                    ast::CallArg::Kwarg(key, value) => {
+                        if static_kwargs {
+                            if let ast::Expr::Const(c) = value {
+                                collected_kwargs.insert(Value::from(*key), c.value.clone());
+                            } else {
+                                unreachable!();
+                            }
+                        } else {
+                            self.add(Instruction::LoadConst(Value::from(*key)));
+                            self.compile_expr(value);
+                            pending_kwargs += 1;
+                        }
+                    }
+                    ast::CallArg::KwargSplat(expr) => {
+                        if pending_kwargs > 0 {
+                            self.add(Instruction::BuildKwargs(pending_kwargs));
+                            num_kwargs_batches += 1;
+                            pending_kwargs = 0;
+                        }
+                        self.compile_expr(expr);
+                        num_kwargs_batches += 1;
+                    }
+                    ast::CallArg::Pos(_) | ast::CallArg::PosSplat(_) => {}
+                }
+            }
+
+            if !collected_kwargs.is_empty() {
+                self.add(Instruction::LoadConst(Kwargs::wrap(collected_kwargs)));
+            } else {
+                // The conditions above guarantee that if we collect static kwargs
+                // we cannot enter this block (single kwargs batch, no caller).
+
+                #[cfg(feature = "macros")]
+                {
+                    if let Some(caller) = caller {
+                        self.add(Instruction::LoadConst(Value::from("caller")));
+                        self.compile_macro_expression(caller);
+                        pending_kwargs += 1
+                    }
+                }
+                if num_kwargs_batches > 0 {
+                    if pending_kwargs > 0 {
+                        self.add(Instruction::BuildKwargs(pending_kwargs));
+                        num_kwargs_batches += 1;
+                    }
+                    self.add(Instruction::MergeKwargs(num_kwargs_batches));
+                } else {
+                    self.add(Instruction::BuildKwargs(pending_kwargs));
+                }
+            }
+            pending_args += 1;
+        }
+
+        if num_args_batches > 0 {
+            if pending_args > 0 {
+                self.add(Instruction::BuildList(Some(pending_args)));
+                num_args_batches += 1;
+            }
+            self.add(Instruction::UnpackLists(num_args_batches));
+            None
         } else {
-            args.len()
+            assert!(pending_args as u16 as usize == pending_args);
+            Some(pending_args as u16)
         }
     }
 
