@@ -21,8 +21,11 @@ pub(crate) fn no_auto_escape(_: &str) -> AutoEscape {
     doc = r" * [`Json`](AutoEscape::Json): `.json`, `.json5`, `.js`, `.yaml`, `.yml`"
 )]
 /// * [`None`](AutoEscape::None): _all others_
+///
+/// Additionally `.j2` as final extension is ignored. So `.html.j2` is the
+/// considered the same as `.html`.
 pub fn default_auto_escape_callback(name: &str) -> AutoEscape {
-    match name.rsplit('.').next() {
+    match name.strip_suffix(".j2").unwrap_or(name).rsplit('.').next() {
         Some("html" | "htm" | "xml") => AutoEscape::Html,
         #[cfg(feature = "json")]
         Some("json" | "json5" | "js" | "yaml" | "yml") => AutoEscape::Json,
@@ -69,9 +72,13 @@ pub(crate) fn get_builtin_filters() -> BTreeMap<Cow<'static, str>, filters::Boxe
         rv.insert("reverse".into(), BoxedFilter::new(filters::reverse));
         rv.insert("trim".into(), BoxedFilter::new(filters::trim));
         rv.insert("join".into(), BoxedFilter::new(filters::join));
+        rv.insert("split".into(), BoxedFilter::new(filters::split));
+        rv.insert("lines".into(), BoxedFilter::new(filters::lines));
         rv.insert("default".into(), BoxedFilter::new(filters::default));
         rv.insert("round".into(), BoxedFilter::new(filters::round));
         rv.insert("abs".into(), BoxedFilter::new(filters::abs));
+        rv.insert("int".into(), BoxedFilter::new(filters::int));
+        rv.insert("float".into(), BoxedFilter::new(filters::float));
         rv.insert("attr".into(), BoxedFilter::new(filters::attr));
         rv.insert("first".into(), BoxedFilter::new(filters::first));
         rv.insert("last".into(), BoxedFilter::new(filters::last));
@@ -80,6 +87,7 @@ pub(crate) fn get_builtin_filters() -> BTreeMap<Cow<'static, str>, filters::Boxe
         rv.insert("sort".into(), BoxedFilter::new(filters::sort));
         rv.insert("d".into(), BoxedFilter::new(filters::default));
         rv.insert("list".into(), BoxedFilter::new(filters::list));
+        rv.insert("string".into(), BoxedFilter::new(filters::string));
         rv.insert("bool".into(), BoxedFilter::new(filters::bool));
         rv.insert("batch".into(), BoxedFilter::new(filters::batch));
         rv.insert("slice".into(), BoxedFilter::new(filters::slice));
@@ -89,6 +97,7 @@ pub(crate) fn get_builtin_filters() -> BTreeMap<Cow<'static, str>, filters::Boxe
         rv.insert("selectattr".into(), BoxedFilter::new(filters::selectattr));
         rv.insert("rejectattr".into(), BoxedFilter::new(filters::rejectattr));
         rv.insert("map".into(), BoxedFilter::new(filters::map));
+        rv.insert("groupby".into(), BoxedFilter::new(filters::groupby));
         rv.insert("unique".into(), BoxedFilter::new(filters::unique));
         rv.insert("pprint".into(), BoxedFilter::new(filters::pprint));
 
@@ -115,17 +124,26 @@ pub(crate) fn get_builtin_tests() -> BTreeMap<Cow<'static, str>, BoxedTest> {
     rv.insert("escaped".into(), is_safe);
     #[cfg(feature = "builtins")]
     {
+        rv.insert("boolean".into(), BoxedTest::new(tests::is_boolean));
         rv.insert("odd".into(), BoxedTest::new(tests::is_odd));
         rv.insert("even".into(), BoxedTest::new(tests::is_even));
+        rv.insert("divisibleby".into(), BoxedTest::new(tests::is_divisibleby));
         rv.insert("number".into(), BoxedTest::new(tests::is_number));
+        rv.insert("integer".into(), BoxedTest::new(tests::is_integer));
+        rv.insert("int".into(), BoxedTest::new(tests::is_integer));
+        rv.insert("float".into(), BoxedTest::new(tests::is_float));
         rv.insert("string".into(), BoxedTest::new(tests::is_string));
         rv.insert("sequence".into(), BoxedTest::new(tests::is_sequence));
+        rv.insert("iterable".into(), BoxedTest::new(tests::is_iterable));
         rv.insert("mapping".into(), BoxedTest::new(tests::is_mapping));
         rv.insert(
             "startingwith".into(),
             BoxedTest::new(tests::is_startingwith),
         );
         rv.insert("endingwith".into(), BoxedTest::new(tests::is_endingwith));
+        rv.insert("lower".into(), BoxedTest::new(tests::is_lower));
+        rv.insert("upper".into(), BoxedTest::new(tests::is_upper));
+        rv.insert("sameas".into(), BoxedTest::new(tests::is_sameas));
 
         // operators
         let is_eq = BoxedTest::new(tests::is_eq);
@@ -176,7 +194,40 @@ pub(crate) fn get_globals() -> BTreeMap<Cow<'static, str>, Value> {
             "debug".into(),
             BoxedFunction::new(functions::debug).to_value(),
         );
+        rv.insert(
+            "namespace".into(),
+            BoxedFunction::new(functions::namespace).to_value(),
+        );
     }
 
     rv
 }
+
+#[cfg(test)]
+mod unit_tests {
+    use super::*;
+
+    use similar_asserts::assert_eq;
+
+    #[test]
+    fn test_default_uto_escape() {
+        assert_eq!(default_auto_escape_callback("foo.html"), AutoEscape::Html);
+        assert_eq!(
+            default_auto_escape_callback("foo.html.j2"),
+            AutoEscape::Html
+        );
+        assert_eq!(default_auto_escape_callback("foo.htm"), AutoEscape::Html);
+        assert_eq!(default_auto_escape_callback("foo.htm.j2"), AutoEscape::Html);
+        assert_eq!(default_auto_escape_callback("foo.txt"), AutoEscape::None);
+        assert_eq!(default_auto_escape_callback("foo.txt.j2"), AutoEscape::None);
+
+        #[cfg(feature = "json")]
+        {
+            assert_eq!(default_auto_escape_callback("foo.yaml"), AutoEscape::Json);
+            assert_eq!(
+                default_auto_escape_callback("foo.json.j2"),
+                AutoEscape::Json
+            );
+        }
+    }
+}
