@@ -6,19 +6,28 @@ use std::sync::Arc;
 use serde::Serialize;
 
 use crate::compiler::codegen::CodeGenerator;
-use crate::compiler::lexer::SyntaxConfig;
+use crate::compiler::instructions::Instructions;
 use crate::compiler::parser::parse_expr;
 use crate::error::{attach_basic_debug_info, Error, ErrorKind};
 use crate::expression::Expression;
 use crate::output::Output;
-use crate::template::{CompiledTemplate, CompiledTemplateRef, Template};
+use crate::template::{CompiledTemplate, CompiledTemplateRef, Template, TemplateConfig};
 use crate::utils::{AutoEscape, BTreeMapKeysDebug, UndefinedBehavior};
 use crate::value::{FunctionArgs, FunctionResult, Value};
 use crate::vm::State;
 use crate::{defaults, filters, functions, tests};
 
-type AutoEscapeFunc = dyn Fn(&str) -> AutoEscape + Sync + Send;
 type FormatterFunc = dyn Fn(&mut Output, &State, &Value) -> Result<(), Error> + Sync + Send;
+type PathJoinFunc = dyn for<'s> Fn(&'s str, &'s str) -> Cow<'s, str> + Sync + Send;
+type UnknownMethodFunc =
+    dyn Fn(&State, &Value, &str, &[Value]) -> Result<Value, Error> + Sync + Send;
+
+/// The maximum recursion in the VM.  Normally each stack frame
+/// adds one to this counter (eg: every time a frame is added).
+/// However in some situations more depth is pushed if the cost
+/// of the stack frame is higher.  Raising this above this limit
+/// requires enabling the `stacker` feature.
+const MAX_RECURSION: usize = 500;
 
 /// An abstraction that holds the engine configuration.
 ///
@@ -41,13 +50,15 @@ pub struct Environment<'source> {
     filters: BTreeMap<Cow<'source, str>, filters::BoxedFilter>,
     tests: BTreeMap<Cow<'source, str>, tests::BoxedTest>,
     globals: BTreeMap<Cow<'source, str>, Value>,
-    default_auto_escape: Arc<AutoEscapeFunc>,
+    path_join_callback: Option<Arc<PathJoinFunc>>,
+    pub(crate) unknown_method_callback: Option<Arc<UnknownMethodFunc>>,
     undefined_behavior: UndefinedBehavior,
     formatter: Arc<FormatterFunc>,
     #[cfg(feature = "debug")]
     debug: bool,
     #[cfg(feature = "fuel")]
     fuel: Option<u64>,
+    recursion_limit: usize,
 }
 
 impl<'source> Default for Environment<'source> {
@@ -76,17 +87,21 @@ impl<'source> Environment<'source> {
     /// [`empty`](Environment::empty) method.
     pub fn new() -> Environment<'source> {
         Environment {
-            templates: TemplateStore::default(),
+            templates: TemplateStore::new(TemplateConfig::new(Arc::new(
+                defaults::default_auto_escape_callback,
+            ))),
             filters: defaults::get_builtin_filters(),
             tests: defaults::get_builtin_tests(),
             globals: defaults::get_globals(),
-            default_auto_escape: Arc::new(defaults::default_auto_escape_callback),
+            path_join_callback: None,
+            unknown_method_callback: None,
             undefined_behavior: UndefinedBehavior::default(),
             formatter: Arc::new(defaults::escape_formatter),
             #[cfg(feature = "debug")]
             debug: cfg!(debug_assertions),
             #[cfg(feature = "fuel")]
             fuel: None,
+            recursion_limit: MAX_RECURSION,
         }
     }
 
@@ -96,17 +111,19 @@ impl<'source> Environment<'source> {
     /// logic for auto escaping configured.
     pub fn empty() -> Environment<'source> {
         Environment {
-            templates: TemplateStore::default(),
+            templates: TemplateStore::new(TemplateConfig::new(Arc::new(defaults::no_auto_escape))),
             filters: Default::default(),
             tests: Default::default(),
             globals: Default::default(),
-            default_auto_escape: Arc::new(defaults::no_auto_escape),
+            path_join_callback: None,
+            unknown_method_callback: None,
             undefined_behavior: UndefinedBehavior::default(),
             formatter: Arc::new(defaults::escape_formatter),
             #[cfg(feature = "debug")]
             debug: cfg!(debug_assertions),
             #[cfg(feature = "fuel")]
             fuel: None,
+            recursion_limit: MAX_RECURSION,
         }
     }
 
@@ -124,7 +141,7 @@ impl<'source> Environment<'source> {
     ///
     /// Note that there are situations where the interface of this method is
     /// too restrictive as you need to hold on to the strings for the lifetime
-    /// of the environment.
+    /// of the environment.  This methodl fails if the template has a syntax error.
     #[cfg_attr(
         feature = "loader",
         doc = "To address this restriction use [`add_template_owned`](Self::add_template_owned)."
@@ -133,7 +150,7 @@ impl<'source> Environment<'source> {
         self.templates.insert(name, source)
     }
 
-    /// Adds a template without without borrowing.
+    /// Adds a template without borrowing.
     ///
     /// This lets you place an owned [`String`] in the environment rather than the
     /// borrowed `&str` without having to worry about lifetimes.
@@ -145,7 +162,7 @@ impl<'source> Environment<'source> {
     /// ```
     ///
     /// **Note**: the name is a bit of a misnomer as this API also allows to borrow too as
-    /// the parameters are actually [`Cow`].
+    /// the parameters are actually [`Cow`].  This method fails if the template has a syntax error.
     #[cfg(feature = "loader")]
     #[cfg_attr(docsrs, doc(cfg(feature = "loader")))]
     pub fn add_template_owned<N, S>(&mut self, name: N, source: S) -> Result<(), Error>
@@ -193,11 +210,124 @@ impl<'source> Environment<'source> {
         self.templates.set_loader(f);
     }
 
+    /// Preserve the trailing newline when rendering templates.
+    ///
+    /// The default is `false`, which causes a single newline, if present, to be
+    /// stripped from the end of the template.
+    ///
+    /// This setting is used whenever a template is loaded into the environment.
+    /// Changing it at a later point only affects future templates loaded.
+    pub fn set_keep_trailing_newline(&mut self, yes: bool) {
+        self.templates
+            .template_config
+            .ws_config
+            .keep_trailing_newline = yes;
+    }
+
+    /// Returns the value of the trailing newline preservation flag.
+    pub fn keep_trailing_newline(&self) -> bool {
+        self.templates
+            .template_config
+            .ws_config
+            .keep_trailing_newline
+    }
+
+    /// Remove the first newline after a block.
+    ///
+    /// If this is set to `true` then the first newline after a block is removed
+    /// (block, not variable tag!). Defaults to `false`.
+    pub fn set_trim_blocks(&mut self, yes: bool) {
+        self.templates.template_config.ws_config.trim_blocks = yes;
+    }
+
+    /// Returns the value of the trim blocks flag.
+    pub fn trim_blocks(&self) -> bool {
+        self.templates.template_config.ws_config.trim_blocks
+    }
+
+    /// Remove leading spaces and tabs from the start of a line to a block.
+    ///
+    /// If this is set to `true` then leading spaces and tabs from the start of a line
+    /// to the block tag are removed.
+    pub fn set_lstrip_blocks(&mut self, yes: bool) {
+        self.templates.template_config.ws_config.lstrip_blocks = yes;
+    }
+
+    /// Returns the value of the lstrip blocks flag.
+    pub fn lstrip_blocks(&self) -> bool {
+        self.templates.template_config.ws_config.lstrip_blocks
+    }
+
     /// Removes a template by name.
     pub fn remove_template(&mut self, name: &str) {
         self.templates.remove(name);
     }
 
+    /// Sets a callback to join template paths.
+    ///
+    /// By default this returns the template path unchanged, but it can be used
+    /// to implement relative path resolution between templates.  The first
+    /// argument to the callback is the name of the template to be loaded, the
+    /// second argument is the parent path.
+    ///
+    /// The following example demonstrates how a basic path joining algorithm
+    /// can be implemented.
+    ///
+    /// ```
+    /// # let mut env = minijinja::Environment::new();
+    /// env.set_path_join_callback(|name, parent| {
+    ///     let mut rv = parent.split('/').collect::<Vec<_>>();
+    ///     rv.pop();
+    ///     name.split('/').for_each(|segment| match segment {
+    ///         "." => {}
+    ///         ".." => { rv.pop(); }
+    ///         _ => { rv.push(segment); }
+    ///     });
+    ///     rv.join("/").into()
+    /// });
+    /// ```
+    pub fn set_path_join_callback<F>(&mut self, f: F)
+    where
+        F: for<'s> Fn(&'s str, &'s str) -> Cow<'s, str> + Send + Sync + 'static,
+    {
+        self.path_join_callback = Some(Arc::new(f));
+    }
+
+    /// Sets a callback invoked for unknown methods on objects.
+    ///
+    /// This registers a function with the environment that is invoked when invoking a method
+    /// on a value results in a [`UnknownMethod`](crate::ErrorKind::UnknownMethod) error.
+    /// In that case the callback is invoked with [`State`], the [`Value`], the name of
+    /// the method as `&str` as well as all arguments in a slice.
+    ///
+    /// This for instance implements a `.items()` method that invokes the `|items` filter:
+    ///
+    /// ```rust
+    /// use minijinja::value::{ValueKind, from_args};
+    /// use minijinja::{Error, ErrorKind};
+    /// # let mut env = minijinja::Environment::new();
+    ///
+    /// env.set_unknown_method_callback(|state, value, method, args| {
+    ///     if value.kind() == ValueKind::Map && method == "items" {
+    ///         let _: () = from_args(args)?;
+    ///         state.apply_filter("items", &[value.clone()])
+    ///     } else {
+    ///         Err(Error::from(ErrorKind::UnknownMethod))
+    ///     }
+    /// });
+    /// ```
+    ///
+    /// This can be used to increase the compatibility with Jinja2 templates that might
+    /// call Python methods on objects which are not available in minijinja.  A range of
+    /// common Python methods is implemented in `minijinja-contrib`.  For more information
+    /// see [minijinja_contrib::pycompat](https://docs.rs/minijinja-contrib/latest/minijinja_contrib/pycompat/).
+    pub fn set_unknown_method_callback<F>(&mut self, f: F)
+    where
+        F: Fn(&State, &Value, &str, &[Value]) -> Result<Value, Error> + Sync + Send + 'static,
+    {
+        self.unknown_method_callback = Some(Arc::new(f));
+    }
+
     /// Removes all stored templates.
     ///
     /// This method is mainly useful when combined with a loader as it causes
@@ -207,11 +337,33 @@ impl<'source> Environment<'source> {
         self.templates.clear();
     }
 
+    /// Returns an iterator over the already loaded templates and their names.
+    ///
+    /// Only templates that are already loaded will be returned.
+    ///
+    /// ```
+    /// # use minijinja::{Environment, context};
+    /// let mut env = Environment::new();
+    /// env.add_template("hello.txt", "Hello {{ name }}!").unwrap();
+    /// env.add_template("goodbye.txt", "Goodbye {{ name }}!").unwrap();
+    ///
+    /// # assert_eq!(env.templates().count(), 2);
+    /// for (name, tmpl) in env.templates() {
+    ///     println!("{}", tmpl.render(context!{ name => "World" }).unwrap());
+    /// }
+    /// ```
+    pub fn templates(&self) -> impl Iterator<Item = (&str, Template<'_, '_>)> {
+        self.templates.iter().map(|(name, template)| {
+            let template = Template::new(self, CompiledTemplateRef::Borrowed(template));
+            (name, template)
+        })
+    }
+
     /// Fetches a template by name.
     ///
     /// This requires that the template has been loaded with
     /// [`add_template`](Environment::add_template) beforehand.  If the template was
-    /// not loaded an error of kind `TemplateNotFound` is returned.  If a loaded was
+    /// not loaded an error of kind `TemplateNotFound` is returned.  If a loader was
     /// added to the engine this can also dynamically load templates.
     ///
     /// ```
@@ -223,11 +375,7 @@ impl<'source> Environment<'source> {
     /// ```
     pub fn get_template(&self, name: &str) -> Result<Template<'_, '_>, Error> {
         let compiled = ok!(self.templates.get(name));
-        Ok(Template::new(
-            self,
-            CompiledTemplateRef::Borrowed(compiled),
-            self.initial_auto_escape(name),
-        ))
+        Ok(Template::new(self, CompiledTemplateRef::Borrowed(compiled)))
     }
 
     /// Loads a template from a string.
@@ -252,9 +400,8 @@ impl<'source> Environment<'source> {
             CompiledTemplateRef::Owned(Arc::new(ok!(CompiledTemplate::new(
                 name,
                 source,
-                self._syntax_config().clone(),
+                &self.templates.template_config,
             )))),
-            self.initial_auto_escape(name),
         ))
     }
 
@@ -306,14 +453,14 @@ impl<'source> Environment<'source> {
     /// **Note on values:** The [`Value`] type implements `Serialize` and can be
     /// efficiently passed to render.  It does not undergo actual serialization.
     pub fn render_str<S: Serialize>(&self, source: &str, ctx: S) -> Result<String, Error> {
-        // reduce total amount of code faling under mono morphization into
+        // reduce total amount of code falling under mono morphization into
         // this function, and share the rest in _eval.
         ok!(self.template_from_str(source)).render(ctx)
     }
 
     /// Sets a new function to select the default auto escaping.
     ///
-    /// This function is invoked when templates are loaded from the environment
+    /// This function is invoked when templates are loaded into the environment
     /// to determine the default auto escaping behavior.  The function is
     /// invoked with the name of the template and can make an initial auto
     /// escaping decision based on that.  The default implementation
@@ -335,7 +482,7 @@ impl<'source> Environment<'source> {
     where
         F: Fn(&str) -> AutoEscape + 'static + Sync + Send,
     {
-        self.default_auto_escape = Arc::new(f);
+        self.templates.template_config.default_auto_escape = Arc::new(f);
     }
 
     /// Changes the undefined behavior.
@@ -448,26 +595,54 @@ impl<'source> Environment<'source> {
 
     /// Sets the syntax for the environment.
     ///
-    /// Note that when `source` is used, the syntax is held on the underlying source
-    /// which means that the actual source needs to have it's syntax changed.
+    /// This setting is used whenever a template is loaded into the environment.
+    /// Changing it at a later point only affects future templates loaded.
     ///
-    /// See [`Syntax`](crate::Syntax) for more information.
+    /// See [`SyntaxConfig`](crate::syntax::SyntaxConfig) for more information.
     #[cfg(feature = "custom_syntax")]
     #[cfg_attr(docsrs, doc(cfg(feature = "custom_syntax")))]
-    pub fn set_syntax(&mut self, syntax: crate::custom_syntax::Syntax) -> Result<(), Error> {
-        self.templates.syntax_config = ok!(syntax.compile());
-        Ok(())
+    pub fn set_syntax(&mut self, syntax: crate::syntax::SyntaxConfig) {
+        self.templates.template_config.syntax_config = syntax;
     }
 
-    /// Returns the current syntax.
+    /// Returns the current syntax config.
     #[cfg(feature = "custom_syntax")]
     #[cfg_attr(docsrs, doc(cfg(feature = "custom_syntax")))]
-    pub fn syntax(&self) -> &crate::custom_syntax::Syntax {
-        &self._syntax_config().syntax
+    pub fn syntax(&self) -> &crate::syntax::SyntaxConfig {
+        &self.templates.template_config.syntax_config
     }
 
-    fn _syntax_config(&self) -> &SyntaxConfig {
-        &self.templates.syntax_config
+    /// Reconfigures the runtime recursion limit.
+    ///
+    /// This defaults to `500`.  Raising it above that level requires the `stacker`
+    /// feature to be enabled.  Otherwise the limit is silently capped at that safe
+    /// maximum.  Note that the maximum is not necessarily safe if the thread uses
+    /// a lot of stack space already, it's just a value that was validated once to
+    /// provide basic protection.
+    ///
+    /// Every operation that requires recursion in MiniJinja increments an internal
+    /// recursion counter.  The actual cost attributed to that recursion depends on
+    /// the cost of the operation.  If statements and for loops for instance only
+    /// increase the counter by 1, whereas template includes and macros might increase
+    /// it to 10 or more.
+    ///
+    /// **Note on stack growth:** even if the stacker feature is enabled it does not
+    /// mean that in all cases stack can grow to the limits desired.  For instance in
+    /// WASM the maximum limits are additionally enforced by the runtime.
+    pub fn set_recursion_limit(&mut self, level: usize) {
+        #[cfg(not(feature = "stacker"))]
+        {
+            self.recursion_limit = level.min(MAX_RECURSION);
+        }
+        #[cfg(feature = "stacker")]
+        {
+            self.recursion_limit = level;
+        }
+    }
+
+    /// Returns the current max recursion limit.
+    pub fn recursion_limit(&self) -> usize {
+        self.recursion_limit
     }
 
     /// Compiles an expression.
@@ -477,15 +652,35 @@ impl<'source> Environment<'source> {
     /// be used as a minimal scripting language.  For more information and an
     /// example see [`Expression`].
     pub fn compile_expression(&self, expr: &'source str) -> Result<Expression<'_, 'source>, Error> {
-        attach_basic_debug_info(self._compile_expression(expr), expr)
+        self._compile_expression(expr)
+            .map(|instr| Expression::new(self, instr))
     }
 
-    fn _compile_expression(&self, expr: &'source str) -> Result<Expression<'_, 'source>, Error> {
-        let ast = ok!(parse_expr(expr, self._syntax_config().clone()));
-        let mut gen = CodeGenerator::new("<expression>", expr);
-        gen.compile_expr(&ast);
-        let (instructions, _) = gen.finish();
-        Ok(Expression::new(self, instructions))
+    /// Compiles an expression without capturing the lifetime.
+    ///
+    /// This works exactly like [`compile_expression`](Self::compile_expression) but
+    /// lets you pass an owned string without capturing the lifetime.
+    #[cfg(feature = "loader")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "loader")))]
+    pub fn compile_expression_owned<E>(&self, expr: E) -> Result<Expression<'_, 'source>, Error>
+    where
+        E: Into<Cow<'source, str>>,
+    {
+        crate::loader::OwnedInstructions::try_new(Box::from(expr.into()), |expr| {
+            self._compile_expression(expr)
+        })
+        .map(|instr| Expression::new_owned(self, instr))
+    }
+
+    fn _compile_expression<'expr>(&self, expr: &'expr str) -> Result<Instructions<'expr>, Error> {
+        attach_basic_debug_info(
+            parse_expr(expr).map(|ast| {
+                let mut gen = CodeGenerator::new("<expression>", expr);
+                gen.compile_expr(&ast);
+                gen.finish().0
+            }),
+            expr,
+        )
     }
 
     /// Adds a new filter function.
@@ -585,7 +780,7 @@ impl<'source> Environment<'source> {
     }
 
     pub(crate) fn initial_auto_escape(&self, name: &str) -> AutoEscape {
-        (self.default_auto_escape)(name)
+        (self.templates.template_config.default_auto_escape)(name)
     }
 
     /// Formats a value into the final format.
@@ -605,15 +800,23 @@ impl<'source> Environment<'source> {
             (self.formatter)(out, state, value)
         }
     }
+
+    /// Performs a template path join.
+    pub(crate) fn join_template_path<'s>(&self, name: &'s str, parent: &'s str) -> Cow<'s, str> {
+        match self.path_join_callback {
+            Some(ref cb) => cb(name, parent),
+            None => Cow::Borrowed(name),
+        }
+    }
 }
 
 #[cfg(not(feature = "loader"))]
 mod basic_store {
     use super::*;
 
-    #[derive(Clone, Default)]
-    pub struct BasicStore<'source> {
-        pub syntax_config: SyntaxConfig,
+    #[derive(Clone)]
+    pub(crate) struct BasicStore<'source> {
+        pub template_config: TemplateConfig,
         map: BTreeMap<&'source str, Arc<CompiledTemplate<'source>>>,
     }
 
@@ -624,13 +827,20 @@ mod basic_store {
     }
 
     impl<'source> BasicStore<'source> {
+        pub fn new(template_config: TemplateConfig) -> BasicStore<'source> {
+            BasicStore {
+                template_config,
+                map: BTreeMap::default(),
+            }
+        }
+
         pub fn insert(&mut self, name: &'source str, source: &'source str) -> Result<(), Error> {
             self.map.insert(
                 name,
                 Arc::new(ok!(CompiledTemplate::new(
                     name,
                     source,
-                    self.syntax_config.clone()
+                    &self.template_config
                 ))),
             );
             Ok(())
@@ -650,6 +860,10 @@ mod basic_store {
                 .map(|x| &**x)
                 .ok_or_else(|| Error::new_not_found(name))
         }
+
+        pub fn iter(&self) -> impl Iterator<Item = (&str, &CompiledTemplate<'source>)> {
+            self.map.iter().map(|(name, template)| (*name, &**template))
+        }
     }
 }
 
