@@ -48,21 +48,58 @@
 //! # Accessing State
 //!
 //! In some cases it can be necessary to access the execution [`State`].  Since a borrowed
-//! state implements [`ArgType`](crate::value::ArgType) it's possible to add a
-//! parameter that holds the state.  For instance the following filter appends
-//! the current template name to the string:
+//! state implements [`ArgType`] it's possible to add a parameter that holds the state.
+//! For instance the following filter appends the current template name to the string:
 //!
 //! ```
 //! # use minijinja::Environment;
 //! # let mut env = Environment::new();
-//! use minijinja::value::Value;
-//! use minijinja::State;
+//! use minijinja::{Value, State};
 //!
 //! fn append_template(state: &State, value: &Value) -> String {
 //!     format!("{}-{}", value, state.name())
 //! }
 //!
-//! env.add_filter("appendTemplate", append_template);
+//! env.add_filter("append_template", append_template);
+//! ```
+//!
+//! # Filter configuration
+//!
+//! The recommended pattern for filters to change their behavior is to leverage global
+//! variables in the template.  For instance take a filter that performs date formatting.
+//! You might want to change the default time format format on a per-template basis
+//! without having to update every filter invocation.  In this case the recommended
+//! pattern is to reserve upper case variables and look them up in the filter:
+//!
+//! ```
+//! # use minijinja::Environment;
+//! # let mut env = Environment::new();
+//! # fn format_unix_timestamp(_: f64, _: &str) -> String { "".into() }
+//! use minijinja::State;
+//!
+//! fn timeformat(state: &State, ts: f64) -> String {
+//!     let configured_format = state.lookup("TIME_FORMAT");
+//!     let format = configured_format
+//!         .as_ref()
+//!         .and_then(|x| x.as_str())
+//!         .unwrap_or("HH:MM:SS");
+//!     format_unix_timestamp(ts, format)
+//! }
+//!
+//! env.add_filter("timeformat", timeformat);
+//! ```
+//!
+//! This then later lets a user override the default either by using
+//! [`add_global`](crate::Environment::add_global) or by passing it with the
+//! [`context!`] macro or similar.
+//!
+//! ```
+//! # use minijinja::context;
+//! # let other_variables = context!{};
+//! let ctx = context! {
+//!     TIME_FORMAT => "HH:MM",
+//!     ..other_variables
+//! };
 //! ```
 //!
 //! # Built-in Filters
@@ -105,7 +142,7 @@ pub(crate) struct BoxedFilter(Arc<FilterFunc>);
 /// applied to and up to 4 extra parameters.  The extra parameters can be
 /// marked optional by using `Option<T>`.  The last argument can also use
 /// [`Rest<T>`](crate::value::Rest) to capture the remaining arguments.  All
-/// types are supported for which [`ArgType`](crate::value::ArgType) is implemented.
+/// types are supported for which [`ArgType`] is implemented.
 ///
 /// For a list of built-in filters see [`filters`](crate::filters).
 ///
@@ -256,7 +293,9 @@ mod builtins {
     use super::*;
 
     use crate::error::ErrorKind;
-    use crate::value::{Kwargs, ValueKind, ValueRepr};
+    use crate::utils::splitn_whitespace;
+    use crate::value::ops::as_f64;
+    use crate::value::{Enumerator, Kwargs, Object, ObjectRepr, ValueKind, ValueRepr};
     use std::borrow::Cow;
     use std::cmp::Ordering;
     use std::fmt::Write;
@@ -355,7 +394,7 @@ mod builtins {
         })
     }
 
-    fn sort_helper(a: &Value, b: &Value, case_sensitive: bool) -> Ordering {
+    fn cmp_helper(a: &Value, b: &Value, case_sensitive: bool) -> Ordering {
         if !case_sensitive {
             if let (Some(a), Some(b)) = (a.as_str(), b.as_str()) {
                 #[cfg(feature = "unicode")]
@@ -402,7 +441,7 @@ mod builtins {
             let case_sensitive = ok!(kwargs.get::<Option<bool>>("case_sensitive")).unwrap_or(false);
             rv.sort_by(|a, b| {
                 let (a, b) = if by_value { (&a.1, &b.1) } else { (&a.0, &b.0) };
-                sort_helper(a, b, case_sensitive)
+                cmp_helper(a, b, case_sensitive)
             });
             if let Some(true) = ok!(kwargs.get("reverse")) {
                 rv.reverse();
@@ -456,7 +495,7 @@ mod builtins {
         }
     }
 
-    /// Reverses a list or string
+    /// Reverses an iterable or string
     ///
     /// ```jinja
     /// {% for user in users|reverse %}
@@ -465,16 +504,7 @@ mod builtins {
     /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn reverse(v: Value) -> Result<Value, Error> {
-        if let Some(s) = v.as_str() {
-            Ok(Value::from(s.chars().rev().collect::<String>()))
-        } else if let Some(seq) = v.as_seq() {
-            Ok(Value::from(seq.iter().rev().collect::<Vec<_>>()))
-        } else {
-            Err(Error::new(
-                ErrorKind::InvalidOperation,
-                format!("cannot reverse value of type {}", v.kind()),
-            ))
-        }
+        v.reverse()
     }
 
     /// Trims a value
@@ -497,35 +527,69 @@ mod builtins {
         }
 
         let joiner = joiner.as_ref().unwrap_or(&Cow::Borrowed(""));
-
-        if let Some(s) = val.as_str() {
-            let mut rv = String::new();
-            for c in s.chars() {
-                if !rv.is_empty() {
-                    rv.push_str(joiner);
-                }
-                rv.push(c);
-            }
-            Ok(rv)
-        } else if let Some(seq) = val.as_seq() {
-            let mut rv = String::new();
-            for item in seq.iter() {
-                if !rv.is_empty() {
-                    rv.push_str(joiner);
-                }
-                if let Some(s) = item.as_str() {
-                    rv.push_str(s);
-                } else {
-                    write!(rv, "{item}").ok();
-                }
-            }
-            Ok(rv)
-        } else {
-            Err(Error::new(
+        let iter = ok!(val.try_iter().map_err(|err| {
+            Error::new(
                 ErrorKind::InvalidOperation,
                 format!("cannot join value of type {}", val.kind()),
-            ))
+            )
+            .with_source(err)
+        }));
+
+        let mut rv = String::new();
+        for item in iter {
+            if !rv.is_empty() {
+                rv.push_str(joiner);
+            }
+            if let Some(s) = item.as_str() {
+                rv.push_str(s);
+            } else {
+                write!(rv, "{item}").ok();
+            }
         }
+        Ok(rv)
+    }
+
+    /// Split a string into its substrings, using `split` as the separator string.
+    ///
+    /// If `split` is not provided or `none` the string is split at all whitespace
+    /// characters and multiple spaces and empty strings will be removed from the
+    /// result.
+    ///
+    /// The `maxsplits` parameter defines the maximum number of splits
+    /// (starting from the left).  Note that this follows Python conventions
+    /// rather than Rust ones so `1` means one split and two resulting items.
+    ///
+    /// ```jinja
+    /// {{ "hello world"|split|list }}
+    ///     -> ["hello", "world"]
+    ///
+    /// {{ "c,s,v"|split(",")|list }}
+    ///     -> ["c", "s", "v"]
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn split(s: Arc<str>, split: Option<Arc<str>>, maxsplits: Option<i64>) -> Value {
+        let maxsplits = maxsplits.and_then(|x| if x >= 0 { Some(x as usize + 1) } else { None });
+
+        Value::make_object_iterable((s, split), move |(s, split)| match (split, maxsplits) {
+            (None, None) => Box::new(s.split_whitespace().map(Value::from)),
+            (Some(split), None) => Box::new(s.split(split as &str).map(Value::from)),
+            (None, Some(n)) => Box::new(splitn_whitespace(s, n).map(Value::from)),
+            (Some(split), Some(n)) => Box::new(s.splitn(n, split as &str).map(Value::from)),
+        })
+    }
+
+    /// Splits a string into lines.
+    ///
+    /// The newline character is removed in the process and not retained.  This
+    /// function supports both Windows and UNIX style newlines.
+    ///
+    /// ```jinja
+    /// {{ "foo\nbar\nbaz"|lines }}
+    ///     -> ["foo", "bar", "baz"]
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn lines(s: Arc<str>) -> Value {
+        Value::from_iter(s.lines().map(|x| x.to_string()))
     }
 
     /// If the value is undefined it will return the passed default value,
@@ -552,16 +616,83 @@ mod builtins {
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn abs(value: Value) -> Result<Value, Error> {
         match value.0 {
-            ValueRepr::I64(x) => Ok(Value::from(x.abs())),
-            ValueRepr::I128(x) => Ok(Value::from(x.0.abs())),
+            ValueRepr::U64(_) | ValueRepr::U128(_) => Ok(value),
+            ValueRepr::I64(x) => match x.checked_abs() {
+                Some(rv) => Ok(Value::from(rv)),
+                None => Ok(Value::from((x as i128).abs())), // this cannot overflow
+            },
+            ValueRepr::I128(x) => {
+                x.0.checked_abs()
+                    .map(Value::from)
+                    .ok_or_else(|| Error::new(ErrorKind::InvalidOperation, "overflow on abs"))
+            }
             ValueRepr::F64(x) => Ok(Value::from(x.abs())),
             _ => Err(Error::new(
                 ErrorKind::InvalidOperation,
-                "cannot round value",
+                "cannot get absolute value",
             )),
         }
     }
 
+    /// Converts a value into an integer.
+    ///
+    /// ```jinja
+    /// {{ "42"|int == 42 }} -> true
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn int(value: Value) -> Result<Value, Error> {
+        match &value.0 {
+            ValueRepr::Undefined | ValueRepr::None => Ok(Value::from(0)),
+            ValueRepr::Bool(x) => Ok(Value::from(*x as u64)),
+            ValueRepr::U64(_) | ValueRepr::I64(_) | ValueRepr::U128(_) | ValueRepr::I128(_) => {
+                Ok(value)
+            }
+            ValueRepr::F64(v) => Ok(Value::from(*v as i128)),
+            ValueRepr::String(..) | ValueRepr::SmallStr(_) => {
+                let s = value.as_str().unwrap();
+                if let Ok(i) = s.parse::<i128>() {
+                    Ok(Value::from(i))
+                } else {
+                    match s.parse::<f64>() {
+                        Ok(f) => Ok(Value::from(f as i128)),
+                        Err(err) => Err(Error::new(ErrorKind::InvalidOperation, err.to_string())),
+                    }
+                }
+            }
+            ValueRepr::Bytes(_) | ValueRepr::Object(_) => Err(Error::new(
+                ErrorKind::InvalidOperation,
+                format!("cannot convert {} to integer", value.kind()),
+            )),
+            ValueRepr::Invalid(_) => value.validate(),
+        }
+    }
+
+    /// Converts a value into a float.
+    ///
+    /// ```jinja
+    /// {{ "42.5"|float == 42.5 }} -> true
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn float(value: Value) -> Result<Value, Error> {
+        match &value.0 {
+            ValueRepr::Undefined | ValueRepr::None => Ok(Value::from(0.0)),
+            ValueRepr::Bool(x) => Ok(Value::from(*x as u64 as f64)),
+            ValueRepr::String(..) | ValueRepr::SmallStr(_) => value
+                .as_str()
+                .unwrap()
+                .parse::<f64>()
+                .map(Value::from)
+                .map_err(|err| Error::new(ErrorKind::InvalidOperation, err.to_string())),
+            ValueRepr::Invalid(_) => value.validate(),
+            _ => as_f64(&value, true).map(Value::from).ok_or_else(|| {
+                Error::new(
+                    ErrorKind::InvalidOperation,
+                    format!("cannot convert {} to float", value.kind()),
+                )
+            }),
+        }
+    }
+
     /// Looks up an attribute.
     ///
     /// In MiniJinja this is the same as the `[]` operator.  In Jinja2 there is a
@@ -588,19 +719,21 @@ mod builtins {
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn round(value: Value, precision: Option<i32>) -> Result<Value, Error> {
         match value.0 {
-            ValueRepr::I64(_) | ValueRepr::I128(_) => Ok(value),
+            ValueRepr::I64(_) | ValueRepr::I128(_) | ValueRepr::U64(_) | ValueRepr::U128(_) => {
+                Ok(value)
+            }
             ValueRepr::F64(val) => {
                 let x = 10f64.powi(precision.unwrap_or(0));
                 Ok(Value::from((x * val).round() / x))
             }
             _ => Err(Error::new(
                 ErrorKind::InvalidOperation,
-                "cannot round value",
+                format!("cannot round value ({})", value.kind()),
             )),
         }
     }
 
-    /// Returns the first item from a list.
+    /// Returns the first item from an iterable.
     ///
     /// If the list is empty `undefined` is returned.
     ///
@@ -614,8 +747,8 @@ mod builtins {
     pub fn first(value: Value) -> Result<Value, Error> {
         if let Some(s) = value.as_str() {
             Ok(s.chars().next().map_or(Value::UNDEFINED, Value::from))
-        } else if let Some(s) = value.as_seq() {
-            Ok(s.get_item(0).unwrap_or(Value::UNDEFINED))
+        } else if let Some(mut iter) = value.as_object().and_then(|x| x.try_iter()) {
+            Ok(iter.next().unwrap_or(Value::UNDEFINED))
         } else {
             Err(Error::new(
                 ErrorKind::InvalidOperation,
@@ -624,7 +757,7 @@ mod builtins {
         }
     }
 
-    /// Returns the last item from a list.
+    /// Returns the last item from an iterable.
     ///
     /// If the list is empty `undefined` is returned.
     ///
@@ -642,9 +775,11 @@ mod builtins {
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn last(value: Value) -> Result<Value, Error> {
         if let Some(s) = value.as_str() {
-            Ok(s.chars().rev().next().map_or(Value::UNDEFINED, Value::from))
-        } else if let Some(seq) = value.as_seq() {
-            Ok(seq.iter().last().unwrap_or(Value::UNDEFINED))
+            Ok(s.chars().next_back().map_or(Value::UNDEFINED, Value::from))
+        } else if matches!(value.kind(), ValueKind::Seq | ValueKind::Iterable) {
+            let rev = ok!(value.reverse());
+            let mut iter = ok!(rev.try_iter());
+            Ok(iter.next().unwrap_or_default())
         } else {
             Err(Error::new(
                 ErrorKind::InvalidOperation,
@@ -653,7 +788,11 @@ mod builtins {
         }
     }
 
-    /// Returns the smallest item from the list.
+    /// Returns the smallest item from an iterable.
+    ///
+    /// ```jinja
+    /// {{ [1, 2, 3, 4]|min }} -> 1
+    /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn min(state: &State, value: Value) -> Result<Value, Error> {
         let iter = ok!(state.undefined_behavior().try_iter(value).map_err(|err| {
@@ -662,7 +801,11 @@ mod builtins {
         Ok(iter.min().unwrap_or(Value::UNDEFINED))
     }
 
-    /// Returns the largest item from the list.
+    /// Returns the largest item from an iterable.
+    ///
+    /// ```jinja
+    /// {{ [1, 2, 3, 4]|max }} -> 4
+    /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn max(state: &State, value: Value) -> Result<Value, Error> {
         let iter = ok!(state.undefined_behavior().try_iter(value).map_err(|err| {
@@ -696,11 +839,11 @@ mod builtins {
         let case_sensitive = ok!(kwargs.get::<Option<bool>>("case_sensitive")).unwrap_or(false);
         if let Some(attr) = ok!(kwargs.get::<Option<&str>>("attribute")) {
             items.sort_by(|a, b| match (a.get_path(attr), b.get_path(attr)) {
-                (Ok(a), Ok(b)) => sort_helper(&a, &b, case_sensitive),
+                (Ok(a), Ok(b)) => cmp_helper(&a, &b, case_sensitive),
                 _ => Ordering::Equal,
             });
         } else {
-            items.sort_by(|a, b| sort_helper(a, b, case_sensitive))
+            items.sort_by(|a, b| cmp_helper(a, b, case_sensitive))
         }
         if let Some(true) = ok!(kwargs.get("reverse")) {
             items.reverse();
@@ -723,10 +866,26 @@ mod builtins {
         Ok(Value::from(iter.collect::<Vec<_>>()))
     }
 
+    /// Converts a value into a string if it's not one already.
+    ///
+    /// If the string has been marked as safe, that value is preserved.
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn string(value: Value) -> Value {
+        if value.kind() == ValueKind::String {
+            value
+        } else {
+            value.to_string().into()
+        }
+    }
+
     /// Converts the value into a boolean value.
     ///
     /// This behaves the same as the if statement does with regards to
     /// handling of boolean values.
+    ///
+    /// ```jinja
+    /// {{ 42|bool }} -> true
+    /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn bool(value: Value) -> bool {
         value.is_true()
@@ -859,11 +1018,40 @@ mod builtins {
     /// </script>
     /// <a href="#" data-info='{{ json_object|tojson }}'>...</a>
     /// ```
+    ///
+    /// The filter takes one argument `indent` (which can also be passed as keyword
+    /// argument for compatibility with Jinja2) which can be set to `true` to enable
+    /// pretty printing or an integer to control the indentation of the pretty
+    /// printing feature.
+    ///
+    /// ```jinja
+    /// <script>
+    ///   const GLOBAL_CONFIG = {{ global_config|tojson(indent=2) }};
+    /// </script>
+    /// ```
     #[cfg_attr(docsrs, doc(cfg(all(feature = "builtins", feature = "json"))))]
     #[cfg(feature = "json")]
-    pub fn tojson(value: Value, pretty: Option<bool>) -> Result<Value, Error> {
-        if pretty.unwrap_or(false) {
-            serde_json::to_string_pretty(&value)
+    pub fn tojson(value: Value, indent: Option<Value>, args: Kwargs) -> Result<Value, Error> {
+        let indent = match indent {
+            Some(indent) => Some(indent),
+            None => ok!(args.get("indent")),
+        };
+        let indent = match indent {
+            None => None,
+            Some(ref val) => match bool::try_from(val.clone()).ok() {
+                Some(true) => Some(2),
+                Some(false) => None,
+                None => Some(ok!(usize::try_from(val.clone()))),
+            },
+        };
+        ok!(args.assert_all_used());
+        if let Some(indent) = indent {
+            let mut out = Vec::<u8>::new();
+            let indentation = " ".repeat(indent);
+            let formatter = serde_json::ser::PrettyFormatter::with_indent(indentation.as_bytes());
+            let mut s = serde_json::Serializer::with_formatter(&mut out, formatter);
+            serde::Serialize::serialize(&value, &mut s)
+                .map(|_| unsafe { String::from_utf8_unchecked(out) })
         } else {
             serde_json::to_string(&value)
         }
@@ -886,7 +1074,7 @@ mod builtins {
         })
     }
 
-    /// indents Value with spaces
+    /// Indents Value with spaces
     ///
     /// The first optional parameter to the filter can be set to `true` to
     /// indent the first line. The parameter defaults to false.
@@ -897,12 +1085,11 @@ mod builtins {
     /// ```jinja
     /// example:
     ///   config:
-    /// {{ global_conifg|indent(2) }} #does not indent first line
-    /// {{ global_config|indent(2,true) }} #indent whole Value with two spaces
-    /// {{ global_config|indent(2,true,true)}} #indent whole Value and all Blank Lines value
+    /// {{ global_conifg|indent(2) }}          # does not indent first line
+    /// {{ global_config|indent(2,true) }}     # indent whole Value with two spaces
+    /// {{ global_config|indent(2,true,true)}} # indent whole Value and all blank lines
     /// ```
     #[cfg_attr(docsrs, doc(cfg(all(feature = "builtins"))))]
-    #[cfg(feature = "builtins")]
     pub fn indent(
         mut value: String,
         width: usize,
@@ -910,14 +1097,17 @@ mod builtins {
         indent_blank_lines: Option<bool>,
     ) -> String {
         fn strip_trailing_newline(input: &mut String) {
-            if let Some(stripped) = input.strip_suffix(&['\r', '\n'][..]) {
-                input.truncate(stripped.len());
+            if input.ends_with('\n') {
+                input.truncate(input.len() - 1);
+            }
+            if input.ends_with('\r') {
+                input.truncate(input.len() - 1);
             }
         }
 
         strip_trailing_newline(&mut value);
-
-        let mut output: String = String::new();
+        let indent_with = " ".repeat(width);
+        let mut output = String::new();
         let mut iterator = value.split('\n');
         if !indent_first_line.unwrap_or(false) {
             output.push_str(iterator.next().unwrap());
@@ -926,10 +1116,10 @@ mod builtins {
         for line in iterator {
             if line.is_empty() {
                 if indent_blank_lines.unwrap_or(false) {
-                    output.push_str(&" ".repeat(width));
+                    output.push_str(&indent_with);
                 }
             } else {
-                output.push_str(format!("{}{}", " ".repeat(width), line).as_str());
+                write!(output, "{}{}", indent_with, line).ok();
             }
             output.push('\n');
         }
@@ -958,11 +1148,14 @@ mod builtins {
 
         if value.kind() == ValueKind::Map {
             let mut rv = String::new();
-            for (idx, k) in ok!(value.try_iter()).enumerate() {
-                if idx > 0 {
+            for k in ok!(value.try_iter()) {
+                let v = ok!(value.get_item(&k));
+                if v.is_none() || v.is_undefined() {
+                    continue;
+                }
+                if !rv.is_empty() {
                     rv.push('&');
                 }
-                let v = ok!(value.get_item(&k));
                 write!(
                     rv,
                     "{}={}",
@@ -976,15 +1169,14 @@ mod builtins {
             match &value.0 {
                 ValueRepr::None | ValueRepr::Undefined => Ok("".into()),
                 ValueRepr::Bytes(b) => Ok(percent_encoding::percent_encode(b, SET).to_string()),
-                ValueRepr::String(s, _) => {
-                    Ok(percent_encoding::utf8_percent_encode(s, SET).to_string())
-                }
+                ValueRepr::String(..) | ValueRepr::SmallStr(_) => Ok(
+                    percent_encoding::utf8_percent_encode(value.as_str().unwrap(), SET).to_string(),
+                ),
                 _ => Ok(percent_encoding::utf8_percent_encode(&value.to_string(), SET).to_string()),
             }
         }
     }
 
-    #[cfg(feature = "builtins")]
     fn select_or_reject(
         state: &State,
         invert: bool,
@@ -1036,7 +1228,6 @@ mod builtins {
     /// {{ [false, null, 42]|select }} -> [42]
     /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    #[cfg(feature = "builtins")]
     pub fn select(
         state: &State,
         value: Value,
@@ -1056,7 +1247,6 @@ mod builtins {
     /// {{ users|selectattr("id", "even") }} -> returns all users with an even id
     /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    #[cfg(feature = "builtins")]
     pub fn selectattr(
         state: &State,
         value: Value,
@@ -1071,7 +1261,6 @@ mod builtins {
     ///
     /// This is the inverse of [`select`].
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    #[cfg(feature = "builtins")]
     pub fn reject(
         state: &State,
         value: Value,
@@ -1091,7 +1280,6 @@ mod builtins {
     /// {{ users|rejectattr("id", "even") }} -> returns all users with an odd id
     /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    #[cfg(feature = "builtins")]
     pub fn rejectattr(
         state: &State,
         value: Value,
@@ -1129,7 +1317,6 @@ mod builtins {
     /// Users on this page: {{ titles|map('lower')|join(', ') }}
     /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    #[cfg(feature = "builtins")]
     pub fn map(
         state: &State,
         value: Value,
@@ -1144,7 +1331,11 @@ mod builtins {
             if !args.is_empty() {
                 return Err(Error::from(ErrorKind::TooManyArguments));
             }
-            let default = ok!(kwargs.get::<Option<Value>>("default"));
+            let default = if kwargs.has("default") {
+                ok!(kwargs.get::<Value>("default"))
+            } else {
+                Value::UNDEFINED
+            };
             for value in ok!(state.undefined_behavior().try_iter(value)) {
                 let sub_val = match attr.as_str() {
                     Some(path) => value.get_path(path),
@@ -1153,17 +1344,13 @@ mod builtins {
                 rv.push(match (sub_val, &default) {
                     (Ok(attr), _) => {
                         if attr.is_undefined() {
-                            if let Some(ref default) = default {
-                                default.clone()
-                            } else {
-                                Value::UNDEFINED
-                            }
+                            default.clone()
                         } else {
                             attr
                         }
                     }
-                    (Err(_), Some(default)) => default.clone(),
-                    (Err(err), None) => return Err(err),
+                    (Err(_), default) if !default.is_undefined() => default.clone(),
+                    (Err(err), _) => return Err(err),
                 });
             }
             ok!(kwargs.assert_all_used());
@@ -1192,41 +1379,188 @@ mod builtins {
         Ok(rv)
     }
 
-    /// Returns a list of unique items from the given iterable.
+    /// Group a sequence of objects by an attribute.
     ///
+    /// The attribute can use dot notation for nested access, like `"address.city"``.
+    /// The values are sorted first so only one group is returned for each unique value.
+    /// The attribute can be passed as first argument or as keyword argument named
+    /// `attribute`.
+    ///
+    /// For example, a list of User objects with a city attribute can be
+    /// rendered in groups. In this example, grouper refers to the city value of
+    /// the group.
+    ///
+    /// ```jinja
+    /// <ul>{% for city, items in users|groupby("city") %}
+    ///   <li>{{ city }}
+    ///   <ul>{% for user in items %}
+    ///     <li>{{ user.name }}
+    ///   {% endfor %}</ul>
+    /// </li>
+    /// {% endfor %}</ul>
+    /// ```
+    ///
+    /// groupby yields named tuples of `(grouper, list)``, which can be used instead
+    /// of the tuple unpacking above.  As such this example is equivalent:
+    ///
+    /// ```jinja
+    /// <ul>{% for group in users|groupby(attribute="city") %}
+    ///   <li>{{ group.grouper }}
+    ///   <ul>{% for user in group.list %}
+    ///     <li>{{ user.name }}
+    ///   {% endfor %}</ul>
+    /// </li>
+    /// {% endfor %}</ul>
+    /// ```
+    ///
+    /// You can specify a default value to use if an object in the list does not
+    /// have the given attribute.
+    ///
+    /// ```jinja
+    /// <ul>{% for city, items in users|groupby("city", default="NY") %}
+    ///   <li>{{ city }}: {{ items|map(attribute="name")|join(", ") }}</li>
+    /// {% endfor %}</ul>
+    /// ```
+    ///
+    /// Like the [`sort`] filter, sorting and grouping is case-insensitive by default.
+    /// The key for each group will have the case of the first item in that group
+    /// of values. For example, if a list of users has cities `["CA", "NY", "ca"]``,
+    /// the "CA" group will have two values.  This can be disabled by passing
+    /// `case_sensitive=True`.
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn groupby(value: Value, attribute: Option<&str>, kwargs: Kwargs) -> Result<Value, Error> {
+        let default = ok!(kwargs.get::<Option<Value>>("default")).unwrap_or_default();
+        let case_sensitive = ok!(kwargs.get::<Option<bool>>("case_sensitive")).unwrap_or(false);
+        let attr = match attribute {
+            Some(attr) => attr,
+            None => ok!(kwargs.get::<&str>("attribute")),
+        };
+        let mut items: Vec<Value> = ok!(value.try_iter()).collect();
+        items.sort_by(|a, b| {
+            let a = a.get_path_or_default(attr, &default);
+            let b = b.get_path_or_default(attr, &default);
+            cmp_helper(&a, &b, case_sensitive)
+        });
+        ok!(kwargs.assert_all_used());
+
+        #[derive(Debug)]
+        pub struct GroupTuple {
+            grouper: Value,
+            list: Vec<Value>,
+        }
+
+        impl Object for GroupTuple {
+            fn repr(self: &Arc<Self>) -> ObjectRepr {
+                ObjectRepr::Seq
+            }
+
+            fn get_value(self: &Arc<Self>, key: &Value) -> Option<Value> {
+                match (key.as_usize(), key.as_str()) {
+                    (Some(0), None) | (None, Some("grouper")) => Some(self.grouper.clone()),
+                    (Some(1), None) | (None, Some("list")) => {
+                        Some(Value::make_object_iterable(self.clone(), |this| {
+                            Box::new(this.list.iter().cloned())
+                                as Box<dyn Iterator<Item = _> + Send + Sync>
+                        }))
+                    }
+                    _ => None,
+                }
+            }
+
+            fn enumerate(self: &Arc<Self>) -> Enumerator {
+                Enumerator::Seq(2)
+            }
+        }
+
+        let mut rv = Vec::new();
+        let mut grouper = None::<Value>;
+        let mut list = Vec::new();
+
+        for item in items {
+            let group_by = item.get_path_or_default(attr, &default);
+            if let Some(ref last_grouper) = grouper {
+                if cmp_helper(last_grouper, &group_by, case_sensitive) != Ordering::Equal {
+                    rv.push(Value::from_object(GroupTuple {
+                        grouper: last_grouper.clone(),
+                        list: std::mem::take(&mut list),
+                    }));
+                }
+            }
+            grouper = Some(group_by);
+            list.push(item);
+        }
+
+        if !list.is_empty() {
+            rv.push(Value::from_object(GroupTuple {
+                grouper: grouper.unwrap(),
+                list,
+            }));
+        }
+
+        Ok(Value::from_object(rv))
+    }
+
     /// Returns a list of unique items from the given iterable.
     ///
     /// ```jinja
-    /// {{ ['foo', 'bar', 'foobar', 'FooBar']|unique|list }}
+    /// {{ ['foo', 'bar', 'foobar', 'foobar']|unique|list }}
     ///   -> ['foo', 'bar', 'foobar']
     /// ```
     ///
     /// The unique items are yielded in the same order as their first occurrence
     /// in the iterable passed to the filter.  The filter will not detect
     /// duplicate objects or arrays, only primitives such as strings or numbers.
+    ///
+    /// Optionally the `attribute` keyword argument can be used to make the filter
+    /// operate on an attribute instead of the value itself.  In this case only
+    /// one city per state would be returned:
+    ///
+    /// ```jinja
+    /// {{ list_of_cities|unique(attribute='state') }}
+    /// ```
+    ///
+    /// Like the [`sort`] filter this operates case-insensitive by default.
+    /// For example, if a list has the US state codes `["CA", "NY", "ca"]``,
+    /// the resulting list will have `["CA", "NY"]`.  This can be disabled by
+    /// passing `case_sensitive=True`.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    #[cfg(feature = "builtins")]
-    pub fn unique(values: Vec<Value>) -> Value {
+    pub fn unique(values: Vec<Value>, kwargs: Kwargs) -> Result<Value, Error> {
         use std::collections::BTreeSet;
 
+        let attr = ok!(kwargs.get::<Option<&str>>("attribute"));
+        let case_sensitive = ok!(kwargs.get::<Option<bool>>("case_sensitive")).unwrap_or(false);
+        ok!(kwargs.assert_all_used());
+
         let mut rv = Vec::new();
         let mut seen = BTreeSet::new();
 
         for item in values {
-            if !seen.contains(&item) {
-                rv.push(item.clone());
-                seen.insert(item);
+            let value_to_compare = if let Some(attr) = attr {
+                item.get_path_or_default(attr, &Value::UNDEFINED)
+            } else {
+                item.clone()
+            };
+            let memorized_value = if case_sensitive {
+                value_to_compare.clone()
+            } else if let Some(s) = value_to_compare.as_str() {
+                Value::from(s.to_lowercase())
+            } else {
+                value_to_compare.clone()
+            };
+
+            if !seen.contains(&memorized_value) {
+                rv.push(item);
+                seen.insert(memorized_value);
             }
         }
 
-        Value::from(rv)
+        Ok(Value::from(rv))
     }
 
     /// Pretty print a variable.
     ///
     /// This is useful for debugging as it better shows what's inside an object.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    #[cfg(feature = "builtins")]
     pub fn pprint(value: &Value) -> String {
         format!("{:#?}", value)
     }
