@@ -105,6 +105,7 @@
 //! - [`tests`]: teaches you how to write custom test functions and to see the list of built-in tests.
 //! - [`functions`]: teaches how to write custom functions and to see the list of built-in functions.
 //! - For auto reloading use the [`minijinja-autoreload`](https://docs.rs/minijinja-autoreload) crate.
+//! - For simpler embedding of templates use the [`minijinja-embed`](https://docs.rs/minijinja-embed) crate.
 //!
 //! Additionally there is an [list of examples](https://github.com/mitsuhiko/minijinja/tree/main/examples)
 //! with many different small example programs on GitHub to explore.
@@ -146,16 +147,23 @@
 //!     the `loop` object no longer exist.  Removing this feature can provide faster
 //!     template execution when a lot of loops are involved.
 //!   - `unicode`: when added unicode identifiers are supported and the `sort`
-//!     filter's case insensitive comparising changes to using unicode and not
+//!     filter's case insensitive comparison changes to using unicode and not
 //!     ASCII rules.  Without this features only ASCII identifiers can be used
 //!     for variable names and attributes.
+//!   - `serde`: enables or disables serde support.  In current versions of MiniJinja
+//!     it's not possible to disable serde but it will become possible.  To prevent
+//!     breakage, MiniJinja warns if this feature is disabled.
 //!
 //! - **Rust Functionality:**
 //!
 //!   - `debug`: if this feature is removed some debug functionality of the engine is
 //!     removed as well.  This mainly affects the quality of error reporting.
 //!   - `deserialization`: when removed this disables deserialization support for
-//!     the [`Value`](crate::value::Value) type.
+//!     the [`Value`] type, removes the `ViaDeserialize` type and the error type
+//!     no longer implements `serde::de::Error`.
+//!   - `std_collections`: if this feature is removed some [`Object`](crate::value::Object)
+//!     implementations for standard library collections are removed.  Only the
+//!     ones needed for the engine to function itself are retained.
 //!
 //! There are some additional features that provide extra functionality:
 //!
@@ -169,11 +177,16 @@
 //! - `json`: When enabled the `tojson` filter is added as builtin filter as well as
 //!   the ability to auto escape via `AutoEscape::Json`.
 //! - `urlencode`: When enabled the `urlencode` filter is added as builtin filter.
+//! - `loop_controls`: enables the `{% break %}` and `{% continue %}` loop control flow
+//!    tags.
 //!
 //! Performance and memory related features:
 //!
+//! - `stacker`: enables automatic stack growth which permits much larger levels of recursion
+//!   at runtime.  This does not affect the maximum recursion at parsing time which is always
+//!   limited.
 //! - `speedups`: enables all speedups, in particular it turns on the `v_htmlescape` dependency
-//!   for faster HTML escapling.
+//!   for faster HTML escaping.
 //! - `key_interning`: if this feature is enabled the automatic string interning in
 //!   the value type is enabled.  This feature used to be turned on by default but
 //!   has negative performance effects in newer versions of MiniJinja since a lot of
@@ -181,9 +194,16 @@
 //!   cuts down on memory usage slightly in certain scenarios by interning all string
 //!   keys used in dynamic map values.
 //!
+//! Internals:
+//!
+//! - `unstable_machinery`: exposes an unstable internal API (no semver guarantees) to parse
+//!   templates and interact with the engine internals.  If you need this functionality, please
+//!   leave some feedback on GitHub.
+//!
 //! </details>
 #![allow(clippy::cognitive_complexity)]
 #![allow(clippy::get_first)]
+#![allow(clippy::default_constructed_unit_structs)]
 #![cfg_attr(docsrs, feature(doc_cfg))]
 #![deny(missing_docs)]
 #![doc(html_logo_url = "https://github.com/mitsuhiko/minijinja/raw/main/artwork/logo-square.png")]
@@ -213,9 +233,6 @@ mod loader;
 #[cfg(feature = "loader")]
 pub use loader::path_loader;
 
-#[cfg(feature = "custom_syntax")]
-mod custom_syntax;
-
 #[cfg(feature = "debug")]
 mod debug;
 
@@ -227,12 +244,29 @@ pub use self::output::Output;
 pub use self::template::Template;
 pub use self::utils::{AutoEscape, HtmlEscape, UndefinedBehavior};
 
-#[cfg(feature = "custom_syntax")]
-pub use self::custom_syntax::Syntax;
+/// Re-export for convenience.
+pub use self::value::Value;
 
 pub use self::macros::__context;
 pub use self::vm::State;
 
+// forwards compatibility
+#[cfg(not(feature = "serde"))]
+const _: () = {
+    #[deprecated(
+        since = "2.0.4",
+        note = "Future versions of MiniJinja will require enabling \
+        the 'serde' feature to use serde types.  To silence this warning \
+        add 'serde' to the least of features of minijinja."
+    )]
+    #[allow(unused)]
+    fn enable_implicit_serde_support() {}
+
+    fn trigger_warning() {
+        enable_implicit_serde_support();
+    }
+};
+
 /// This module gives access to the low level machinery.
 ///
 /// This module is only provided by the `unstable_machinery` feature and does not
@@ -245,14 +279,21 @@ pub mod machinery {
     pub use crate::compiler::ast;
     pub use crate::compiler::codegen::CodeGenerator;
     pub use crate::compiler::instructions::{Instruction, Instructions};
-    pub use crate::compiler::lexer::{tokenize, SyntaxConfig};
-    pub use crate::compiler::parser::{parse, parse_with_syntax};
+    pub use crate::compiler::lexer::{tokenize, Tokenizer, WhitespaceConfig};
+    pub use crate::compiler::parser::{parse, parse_expr};
     pub use crate::compiler::tokens::{Span, Token};
-    pub use crate::template::CompiledTemplate;
+    pub use crate::template::{CompiledTemplate, TemplateConfig};
     pub use crate::vm::Vm;
 
     use crate::Output;
 
+    /// Returns a reference to a [`CompiledTemplate`] from a [`Template`](crate::Template).
+    pub fn get_compiled_template<'x, 'env>(
+        tmpl: &'x crate::Template<'env, 'env>,
+    ) -> &'x CompiledTemplate<'env> {
+        &tmpl.compiled
+    }
+
     /// Creates an [`Output`] that writes into a string.
     pub fn make_string_output(s: &mut String) -> Output<'_> {
         Output::with_string(s)
