@@ -1,4 +1,11 @@
-//! Documents the syntax for templates.
+#![cfg_attr(
+    feature = "custom_syntax",
+    doc = "Documents the syntax for templates and provides ways to reconfigure it."
+)]
+#![cfg_attr(
+    not(feature = "custom_syntax"),
+    doc = "Documents the syntax for templates."
+)]
 //!
 //! <details><summary><strong style="cursor: pointer">Table of Contents</strong></summary>
 //!
@@ -26,7 +33,12 @@
 //!   - [`{% do %}`](#-do-)
 //!   - [`{% autoescape %}`](#-autoescape-)
 //!   - [`{% raw %}`](#-raw-)
-//! - [Custom Delimiters](#custom-delimiters)
+//!   - [`{% break %} / {% continue %}`](#-break----continue-)
+#![cfg_attr(
+    feature = "custom_syntax",
+    doc = "- [Custom Delimiters](#custom-delimiters)"
+)]
+//! - [Whitespace Control](#whitespace-control)
 //!
 //! </details>
 //!
@@ -80,8 +92,11 @@
 //! - `"Hello World"`: Everything between two double or single quotes is a string. They are
 //!   useful whenever you need a string in the template (e.g. as arguments to function calls
 //!   and filters, or just to extend or include a template).
-//! - `42`: Integers are whole numbers without a decimal part.
+//! - `42`: Integers are whole numbers without a decimal part.  They can be prefixed with
+//!   `0b` to indicate binary, `0o` to indicate octal and `0x` to indicate hexadecimal.
+//!   Underscores are tolerated (and ignored) everywhere a digit is except in the last place.
 //! - `42.0`: Floating point numbers can be written using a `.` as a decimal mark.
+//!   Underscores are tolerated (and ignored) everywhere a digit is except in the last place.
 //! - `['list', 'of', 'objects']`: Everything between two brackets is a list. Lists are useful
 //!   for storing sequential data to be iterated over.
 //!   for compatibility with Jinja2 `('list', 'of', 'objects')` is also allowed.
@@ -211,6 +226,11 @@
 //! - `loop.changed(...args)`: Returns true if the passed values have changed since the last time it was called with the same arguments.
 //! - `loop.cycle(...args)`: Returns a value from the passed sequence in a cycle.
 //!
+//! A special note on iterators: in the current version of MiniJinja, some sequences are actually
+//! lazy iterators.  They behave a bit like sequences not not entirely.  They can be iterated over,
+//! will happily serialize once into a a list etc.  However when iterating over an actual iterator,
+//! `last`, `revindex` and `revindex0` will always be undefined.
+//!
 //! Within a for-loop, it’s possible to cycle among a list of strings/variables each time through
 //! the loop by using the special `loop.cycle` helper:
 //!
@@ -395,7 +415,7 @@
 //! **Feature:** `multi_template` (included by default)
 //!  
 //! The `include` tag is useful to include a template and return the rendered contents of that file
-//! into the current namespace::
+//! into the current namespace:
 //!  
 //! ```jinja
 //! {% include 'header.html' %}
@@ -504,6 +524,24 @@
 //! {% set title | upper %}Title of the page{% endset %}
 //! ```
 //!
+//! More complex use cases can be handled using namespace objects which allow
+//! propagating of changes across scopes:
+//!
+//! ```jinja
+//! {% set ns = namespace(found=false) %}
+//! {% for item in items %}
+//!   {% if item.check_something() %}
+//!     {% set ns.found = true %}
+//!   {% endif %}
+//!   * {{ item.title }}
+//! {% endfor %}
+//! Found item having something: {{ ns.found }}
+//! ```
+//!
+//! Note that the `obj.attr` notation in the set tag is only allowed for namespace
+//! objects; attempting to assign an attribute on any other object will cause
+//! an error.
+//!
 //! ## `{% filter %}`
 //!
 //! Filter sections allow you to apply regular [filters](crate::filters) on a
@@ -684,29 +722,52 @@
 //! {% endraw %}
 //! ```
 //!
+//! ## `{% break %}` / `{% continue %}`
+//!
+//! If MiniJinja was compiled with the `loop_controls` feature, it’s possible to
+//! use `break`` and `continue`` in loops.  When break is reached, the loop is
+//! terminated; if continue is reached, the processing is stopped and continues
+//! with the next iteration.
+//!
+//! Here’s a loop that skips every second item:
+//!
+//! ```jinja
+//! {% for user in users %}
+//! {%- if loop.index is even %}{% continue %}{% endif %}
+//! ...
+//! {% endfor %}
+//! ```
+//!
+//! Likewise, a loop that stops processing after the 10th iteration:
+//!
+//! ```jinja
+//! {% for user in users %}
+//! {%- if loop.index >= 10 %}{% break %}{% endif %}
+//! {%- endfor %}
+//! ```
+//!
 #![cfg_attr(
     feature = "custom_syntax",
-    doc = r#"
+    doc = r###"
 # Custom Delimiters
 
 When MiniJinja has been compiled with the `custom_syntax` feature (see
-[`Syntax`](crate::Syntax)), it's possible to reconfigure the delimiters of the
+[`SyntaxConfig`]), it's possible to reconfigure the delimiters of the
 templates.  This is generally not recommended but it's useful for situations
 where Jinja templates are used to generate files with a syntax that would be
 conflicting for Jinja.  With custom delimiters it can for instance be more
 convenient to generate LaTeX files:
 
 ```
-# use minijinja::{Environment, Syntax};
+# use minijinja::{Environment, syntax::SyntaxConfig};
 let mut environment = Environment::new();
-environment.set_syntax(Syntax {
-    block_start: "\\BLOCK{".into(),
-    block_end: "}".into(),
-    variable_start: "\\VAR{".into(),
-    variable_end: "}".into(),
-    comment_start: "\\#{".into(),
-    comment_end: "}".into(),
-}).unwrap();
+environment.set_syntax(SyntaxConfig::builder()
+    .block_delimiters("\\BLOCK{", "}")
+    .variable_delimiters("\\VAR{", "}")
+    .comment_delimiters("\\#{", "}")
+    .build()
+    .unwrap()
+);
 ```
 
 And then a template might look like this instead:
@@ -718,5 +779,429 @@ And then a template might look like this instead:
 \BLOCK{endfor}
 \end{itemize}
 ```
-"#
+
+# Line Statements and Comments
+
+MiniJinja supports line statements and comments like Jinja2 does.  Line statements
+and comments are an alternative syntax feature where blocks can be placed on their
+own line if they are opened with a configured prefix.  They must appear on their own
+line but can be prefixed with whitespace.  Line comments are similar but they can
+also be trailing.  These syntax features need to be configured explicitly.
+There are however small differences with regards to whitespace compared to
+Jinja2.
+
+To use line statements and comments the `custom_syntax` feature needs to be
+enabled and they need to be configured (see [`SyntaxConfig`]).
+
+```
+# use minijinja::{Environment, syntax::SyntaxConfig};
+let mut environment = Environment::new();
+environment.set_syntax(SyntaxConfig::builder()
+    .line_statement_prefix("#")
+    .line_comment_prefix("##")
+    .build()
+    .unwrap()
+);
+```
+
+With the above config you can render a template like this:
+
+```jinja
+## This block is
+## completely removed
+<ul>
+  # for item in [1, 2]
+    ## this is a comment that is also removed including leading whitespace
+    <li>{{ item }}
+  # endfor
+</ul>
+```
+
+Renders into the following HTML:
+
+```html
+<ul>
+    <li>1
+    <li>2
+</ul>
+```
+
+Note that this is slightly different than in Jinja2.  Specifically the following
+rules apply with regards to whitespace:
+
+* line statements remove all whitespace before and after including the newline
+* line comments remove the trailing newline and leading whitespace.
+
+This is different than in Jinja2 where empty lines can remain from line comments.
+Additionally whitespace control is not available for line statements.
+"###
 )]
+//! # Whitespace Control
+//!
+//! MiniJinja shares the same behavior with Jinja2 when it comes to
+//! whitespace handling.  By default a single trailing newline is stripped if present
+//! and all other whitespace is returned unchanged.
+//!
+//! If an application configures Jinja to [`trim_blocks`](crate::Environment::set_trim_blocks),
+//! the first newline after a template tag is removed automatically (like in PHP).  The
+//! [`lstrip_blocks`](crate::Environment::set_lstrip_blocks) option can also be set to strip
+//! tabs and spaces from the beginning of a line to the start of a block. (Nothing will be
+//! stripped if there are other characters before the start of the block.)
+//!
+//! With both `trim_blocks` and `lstrip_blocks` enabled, you can put block tags on their
+//! own lines, and the entire block line will be removed when rendered, preserving the
+//! whitespace of the contents.
+//!
+//! For example, without the `trim_blocks` and `lstrip_blocks` options, this template:
+//!
+//! ```jinja
+//! <div>
+//!   {% if True %}
+//!     yay
+//!   {% endif %}
+//! </div>
+//! ```
+//!
+//! gets rendered with blank lines inside the div:
+//!
+//! ```jinja
+//! <div>
+//!
+//!     yay
+//!
+//! </div>
+//! ```
+//!
+//! But with both `trim_blocks` and `lstrip_blocks` enabled, the template block lines
+//! are removed and other whitespace is preserved:
+//!
+//! ```jinja
+//! <div>
+//!     yay
+//! </div>
+//! ````
+//!
+//! You can manually disable the `lstrip_blocks` behavior by putting a plus sign (`+`)
+//! at the start of a block:
+//!
+//! ```jinja
+//! <div>
+//!   {%+ if something %}yay{% endif %}
+//! </div>
+//! ```
+//!
+//! Similarly, you can manually disable the `trim_blocks` behavior by putting a plus
+//! sign (`+`) at the end of a block:
+//!
+//! ```jinja
+//! <div>
+//! {% if something +%}
+//!     yay
+//! {% endif %}
+//! </div>
+//! ```
+//!
+//! You can also strip whitespace in templates by hand. If you add a minus sign (`-`) to the
+//! start or end of a block (e.g. a for tag), a comment, or a variable expression, the
+//! whitespaces before or after that block will be removed:
+//!
+//! ```jinja
+//! {% for item in range(1, 10) -%}
+//! {{ item }}
+//! {%- endfor %}
+//! ```
+//!
+//! This will yield all elements without whitespace between them, in this case
+//! the output would be `123456789`.
+//!
+//! By default, MiniJinja also removes trailing newlines.  To keep single trailing newlines,
+//! configure MiniJinja to [`keep_trailing_newline`](crate::Environment::set_keep_trailing_newline).
+
+#[cfg(feature = "custom_syntax")]
+mod imp {
+    use crate::compiler::lexer::StartMarker;
+    use crate::error::{Error, ErrorKind};
+    use aho_corasick::{AhoCorasick, PatternID};
+    use std::borrow::Cow;
+    use std::sync::Arc;
+
+    #[derive(Debug, PartialEq, Clone)]
+    pub(crate) struct Delims {
+        block_start: Cow<'static, str>,
+        block_end: Cow<'static, str>,
+        variable_start: Cow<'static, str>,
+        variable_end: Cow<'static, str>,
+        comment_start: Cow<'static, str>,
+        comment_end: Cow<'static, str>,
+        line_statement_prefix: Cow<'static, str>,
+        line_comment_prefix: Cow<'static, str>,
+    }
+
+    const DEFAULT_DELIMS: Delims = Delims {
+        block_start: Cow::Borrowed("{%"),
+        block_end: Cow::Borrowed("%}"),
+        variable_start: Cow::Borrowed("{{"),
+        variable_end: Cow::Borrowed("}}"),
+        comment_start: Cow::Borrowed("{#"),
+        comment_end: Cow::Borrowed("#}"),
+        line_statement_prefix: Cow::Borrowed(""),
+        line_comment_prefix: Cow::Borrowed(""),
+    };
+
+    impl Delims {
+        fn validated_start_delims(&self) -> Result<Vec<&str>, Error> {
+            let mut delims = Vec::with_capacity(5);
+            for (delim, required) in [
+                (&self.variable_start, true),
+                (&self.block_start, true),
+                (&self.comment_start, true),
+                (&self.line_statement_prefix, false),
+                (&self.line_comment_prefix, false),
+            ] {
+                let delim = delim as &str;
+                if delim.is_empty() {
+                    if required {
+                        return Err(ErrorKind::InvalidDelimiter.into());
+                    }
+                } else if delims.contains(&delim) {
+                    return Err(ErrorKind::InvalidDelimiter.into());
+                } else {
+                    delims.push(delim);
+                }
+            }
+
+            Ok(delims)
+        }
+    }
+
+    /// Builder helper to reconfigure the syntax.
+    ///
+    /// A new builder is returned by [`SyntaxConfig::builder`].
+    #[derive(Debug)]
+    #[cfg_attr(docsrs, doc(cfg(feature = "custom_syntax")))]
+    pub struct SyntaxConfigBuilder {
+        delims: Arc<Delims>,
+    }
+
+    impl SyntaxConfigBuilder {
+        /// Sets the block start and end delimiters.
+        pub fn block_delimiters<S, E>(&mut self, s: S, e: E) -> &mut Self
+        where
+            S: Into<Cow<'static, str>>,
+            E: Into<Cow<'static, str>>,
+        {
+            let delims = Arc::make_mut(&mut self.delims);
+            delims.block_start = s.into();
+            delims.block_end = e.into();
+            self
+        }
+
+        /// Sets the variable start and end delimiters.
+        pub fn variable_delimiters<S, E>(&mut self, s: S, e: E) -> &mut Self
+        where
+            S: Into<Cow<'static, str>>,
+            E: Into<Cow<'static, str>>,
+        {
+            let delims = Arc::make_mut(&mut self.delims);
+            delims.variable_start = s.into();
+            delims.variable_end = e.into();
+            self
+        }
+
+        /// Sets the comment start and end delimiters.
+        pub fn comment_delimiters<S, E>(&mut self, s: S, e: E) -> &mut Self
+        where
+            S: Into<Cow<'static, str>>,
+            E: Into<Cow<'static, str>>,
+        {
+            let delims = Arc::make_mut(&mut self.delims);
+            delims.comment_start = s.into();
+            delims.comment_end = e.into();
+            self
+        }
+
+        /// Sets the line statement prefix.
+        ///
+        /// By default this is the empty string which disables the line
+        /// statement prefix feature.
+        pub fn line_statement_prefix<S>(&mut self, s: S) -> &mut Self
+        where
+            S: Into<Cow<'static, str>>,
+        {
+            let delims = Arc::make_mut(&mut self.delims);
+            delims.line_statement_prefix = s.into();
+            self
+        }
+
+        /// Sets the line comment prefix.
+        ///
+        /// By default this is the empty string which disables the line
+        /// comment prefix feature.
+        pub fn line_comment_prefix<S>(&mut self, s: S) -> &mut Self
+        where
+            S: Into<Cow<'static, str>>,
+        {
+            let delims = Arc::make_mut(&mut self.delims);
+            delims.line_comment_prefix = s.into();
+            self
+        }
+
+        /// Builds the final syntax config.
+        pub fn build(&self) -> Result<SyntaxConfig, Error> {
+            let delims = self.delims.clone();
+            if *delims == DEFAULT_DELIMS {
+                return Ok(SyntaxConfig::default());
+            }
+            let aho_corasick = ok!(AhoCorasick::builder()
+                .build(ok!(delims.validated_start_delims()))
+                .map_err(|_| ErrorKind::InvalidDelimiter.into()));
+            Ok(SyntaxConfig {
+                delims,
+                aho_corasick: Some(aho_corasick),
+            })
+        }
+    }
+
+    /// The delimiter configuration for the environment and the parser.
+    ///
+    /// Custom syntax configurations require the `custom_syntax` feature.
+    /// Otherwise just the default syntax is available.
+    ///
+    /// You can  override the syntax configuration for templates by setting different
+    /// delimiters.  The end markers can be shared, but the start markers need to be
+    /// distinct.  It would thus not be valid to configure `{{` to be the marker for
+    /// both variables and blocks.
+    ///
+    /// ```
+    /// # use minijinja::{Environment, syntax::SyntaxConfig};
+    /// let mut environment = Environment::new();
+    /// environment.set_syntax(
+    ///     SyntaxConfig::builder()
+    ///         .block_delimiters("\\BLOCK{", "}")
+    ///         .variable_delimiters("\\VAR{", "}")
+    ///         .comment_delimiters("\\#{", "}")
+    ///         .build()
+    ///         .unwrap()
+    /// );
+    #[derive(Clone, Debug)]
+    pub struct SyntaxConfig {
+        delims: Arc<Delims>,
+        pub(crate) aho_corasick: Option<aho_corasick::AhoCorasick>,
+    }
+
+    impl Default for SyntaxConfig {
+        fn default() -> Self {
+            Self {
+                delims: Arc::new(DEFAULT_DELIMS),
+                aho_corasick: None,
+            }
+        }
+    }
+
+    impl SyntaxConfig {
+        /// Creates a syntax builder.
+        #[cfg_attr(docsrs, doc(cfg(feature = "custom_syntax")))]
+        pub fn builder() -> SyntaxConfigBuilder {
+            SyntaxConfigBuilder {
+                delims: Arc::new(DEFAULT_DELIMS),
+            }
+        }
+
+        /// Returns the block delimiters.
+        #[inline(always)]
+        pub fn block_delimiters(&self) -> (&str, &str) {
+            (&self.delims.block_start, &self.delims.block_end)
+        }
+
+        /// Returns the variable delimiters.
+        #[inline(always)]
+        pub fn variable_delimiters(&self) -> (&str, &str) {
+            (&self.delims.variable_start, &self.delims.variable_end)
+        }
+
+        /// Returns the comment delimiters.
+        #[inline(always)]
+        pub fn comment_delimiters(&self) -> (&str, &str) {
+            (&self.delims.comment_start, &self.delims.comment_end)
+        }
+
+        /// Returns the line statement prefix.
+        #[inline(always)]
+        pub fn line_statement_prefix(&self) -> Option<&str> {
+            if self.delims.line_statement_prefix.is_empty() {
+                None
+            } else {
+                Some(&self.delims.line_statement_prefix)
+            }
+        }
+
+        /// Returns the line comment prefix.
+        #[inline(always)]
+        pub fn line_comment_prefix(&self) -> Option<&str> {
+            if self.delims.line_comment_prefix.is_empty() {
+                None
+            } else {
+                Some(&self.delims.line_comment_prefix)
+            }
+        }
+
+        /// Reverse resolves the pattern to a start marker.
+        pub(crate) fn pattern_to_marker(&self, pattern: PatternID) -> StartMarker {
+            match pattern.as_usize() {
+                0 => StartMarker::Variable,
+                1 => StartMarker::Block,
+                2 => StartMarker::Comment,
+                3 => {
+                    if self.line_statement_prefix().is_some() {
+                        StartMarker::LineStatement
+                    } else {
+                        StartMarker::LineComment
+                    }
+                }
+                4 => StartMarker::LineComment,
+                _ => unreachable!(),
+            }
+        }
+    }
+}
+
+#[cfg(not(feature = "custom_syntax"))]
+mod imp {
+    /// The default delimiter configuration for the environment and the parser.
+    #[derive(Clone, Default, Debug)]
+    pub struct SyntaxConfig;
+
+    impl SyntaxConfig {
+        /// Returns the block delimiters.
+        #[inline(always)]
+        pub fn block_delimiters(&self) -> (&str, &str) {
+            ("{%", "%}")
+        }
+
+        /// Returns the variable delimiters.
+        #[inline(always)]
+        pub fn variable_delimiters(&self) -> (&str, &str) {
+            ("{{", "}}")
+        }
+
+        /// Returns the comment delimiters.
+        #[inline(always)]
+        pub fn comment_delimiters(&self) -> (&str, &str) {
+            ("{#", "#}")
+        }
+
+        /// Returns the line statement prefix.
+        #[inline(always)]
+        pub fn line_statement_prefix(&self) -> Option<&str> {
+            None
+        }
+
+        /// Returns the line comment prefix.
+        #[inline(always)]
+        pub fn line_comment_prefix(&self) -> Option<&str> {
+            None
+        }
+    }
+}
+
+pub use self::imp::*;
