@@ -7,15 +7,40 @@ use serde::Serialize;
 
 use crate::compiler::codegen::CodeGenerator;
 use crate::compiler::instructions::Instructions;
-use crate::compiler::lexer::SyntaxConfig;
+use crate::compiler::lexer::WhitespaceConfig;
 use crate::compiler::meta::find_undeclared;
-use crate::compiler::parser::parse_with_syntax;
+use crate::compiler::parser::parse;
 use crate::environment::Environment;
 use crate::error::{attach_basic_debug_info, Error};
 use crate::output::{Output, WriteWrapper};
+use crate::syntax::SyntaxConfig;
 use crate::utils::AutoEscape;
 use crate::value::{self, Value};
-use crate::vm::{prepare_blocks, State, Vm};
+use crate::vm::{prepare_blocks, Context, State, Vm};
+
+/// Callback for auto escape determination
+pub type AutoEscapeFunc = dyn Fn(&str) -> AutoEscape + Sync + Send;
+
+/// Internal struct that holds template loading level config values.
+#[derive(Clone)]
+pub struct TemplateConfig {
+    /// The syntax used for the template.
+    pub syntax_config: SyntaxConfig,
+    /// Controls whitespace behavior.
+    pub ws_config: WhitespaceConfig,
+    /// The callback that determines the initial auto escaping for templates.
+    pub default_auto_escape: Arc<AutoEscapeFunc>,
+}
+
+impl TemplateConfig {
+    pub(crate) fn new(default_auto_escape: Arc<AutoEscapeFunc>) -> TemplateConfig {
+        TemplateConfig {
+            syntax_config: SyntaxConfig::default(),
+            ws_config: WhitespaceConfig::default(),
+            default_auto_escape,
+        }
+    }
+}
 
 /// Represents a handle to a template.
 ///
@@ -27,8 +52,7 @@ use crate::vm::{prepare_blocks, State, Vm};
 #[derive(Clone)]
 pub struct Template<'env: 'source, 'source> {
     env: &'env Environment<'env>,
-    compiled: CompiledTemplateRef<'env, 'source>,
-    initial_auto_escape: AutoEscape,
+    pub(crate) compiled: CompiledTemplateRef<'env, 'source>,
 }
 
 impl<'env, 'source> fmt::Debug for Template<'env, 'source> {
@@ -40,7 +64,7 @@ impl<'env, 'source> fmt::Debug for Template<'env, 'source> {
             ds.field("instructions", &self.compiled.instructions);
             ds.field("blocks", &self.compiled.blocks);
         }
-        ds.field("initial_auto_escape", &self.initial_auto_escape);
+        ds.field("initial_auto_escape", &self.compiled.initial_auto_escape);
         ds.finish()
     }
 }
@@ -49,13 +73,8 @@ impl<'env, 'source> Template<'env, 'source> {
     pub(crate) fn new(
         env: &'env Environment<'env>,
         compiled: CompiledTemplateRef<'env, 'source>,
-        initial_auto_escape: AutoEscape,
     ) -> Template<'env, 'source> {
-        Template {
-            env,
-            compiled,
-            initial_auto_escape,
-        }
+        Template { env, compiled }
     }
 
     /// Returns the name of the template.
@@ -77,8 +96,8 @@ impl<'env, 'source> Template<'env, 'source> {
     ///
     /// For very large contexts and to avoid the overhead of serialization of
     /// potentially unused values, you might consider using a dynamic
-    /// [`StructObject`](crate::value::StructObject) as value.  For more
-    /// information see [Struct as Context](crate::value::StructObject#struct-as-context).
+    /// [`Object`](crate::value::Object) as value.  For more
+    /// information see [Map as Context](crate::value::Object#map-as-context).
     ///
     /// ```
     /// # use minijinja::{Environment, context};
@@ -96,7 +115,7 @@ impl<'env, 'source> Template<'env, 'source> {
     pub fn render<S: Serialize>(&self, ctx: S) -> Result<String, Error> {
         // reduce total amount of code faling under mono morphization into
         // this function, and share the rest in _render.
-        self._render(Value::from_serializable(&ctx)).map(|x| x.0)
+        self._render(Value::from_serialize(&ctx)).map(|x| x.0)
     }
 
     /// Like [`render`](Self::render) but also return the evaluated [`State`].
@@ -122,7 +141,7 @@ impl<'env, 'source> Template<'env, 'source> {
     ) -> Result<(String, State<'_, 'env>), Error> {
         // reduce total amount of code faling under mono morphization into
         // this function, and share the rest in _render.
-        self._render(Value::from_serializable(&ctx))
+        self._render(Value::from_serialize(&ctx))
     }
 
     fn _render(&self, root: Value) -> Result<(String, State<'_, 'env>), Error> {
@@ -156,7 +175,7 @@ impl<'env, 'source> Template<'env, 'source> {
     ) -> Result<State<'_, 'env>, Error> {
         let mut wrapper = WriteWrapper { w, err: None };
         self._eval(
-            Value::from_serializable(&ctx),
+            Value::from_serialize(&ctx),
             &mut Output::with_write(&mut wrapper),
         )
         .map(|(_, state)| state)
@@ -168,18 +187,16 @@ impl<'env, 'source> Template<'env, 'source> {
     /// This evaluates the template, discards the output and returns the final
     /// `State` for introspection.  From there global variables or blocks
     /// can be accessed.  What this does is quite similar to how the engine
-    /// interally works with tempaltes that are extended or imported from.
+    /// internally works with templates that are extended or imported from.
     ///
     /// ```
     /// # use minijinja::{Environment, context};
     /// # fn test() -> Result<(), minijinja::Error> {
     /// # let mut env = Environment::new();
-    /// # env.add_template("hello", "{% block hi %}Hello {{ name }}!{% endblock %}")?;
+    /// # env.add_template("hello", "")?;
     /// let tmpl = env.get_template("hello")?;
-    /// let rv = tmpl
-    ///     .eval_to_state(context!(name => "John"))?
-    ///     .render_block("hi")?;
-    /// println!("{}", rv);
+    /// let state = tmpl.eval_to_state(context!(name => "John"))?;
+    /// println!("{:?}", state.exports());
     /// # Ok(()) }
     /// ```
     ///
@@ -187,7 +204,7 @@ impl<'env, 'source> Template<'env, 'source> {
     ///
     /// For more information see [`State`].
     pub fn eval_to_state<S: Serialize>(&self, ctx: S) -> Result<State<'_, 'env>, Error> {
-        let root = Value::from_serializable(&ctx);
+        let root = Value::from_serialize(&ctx);
         let mut out = Output::null();
         let vm = Vm::new(self.env);
         let state = ok!(vm.eval(
@@ -195,7 +212,7 @@ impl<'env, 'source> Template<'env, 'source> {
             root,
             &self.compiled.blocks,
             &mut out,
-            self.initial_auto_escape,
+            self.compiled.initial_auto_escape,
         ))
         .1;
         Ok(state)
@@ -211,7 +228,7 @@ impl<'env, 'source> Template<'env, 'source> {
             root,
             &self.compiled.blocks,
             out,
-            self.initial_auto_escape,
+            self.compiled.initial_auto_escape,
         )
     }
 
@@ -235,10 +252,12 @@ impl<'env, 'source> Template<'env, 'source> {
     /// // returns ["foo", "bar.baz"]
     /// ```
     pub fn undeclared_variables(&self, nested: bool) -> HashSet<String> {
-        match parse_with_syntax(
+        match parse(
             self.compiled.instructions.source(),
             self.name(),
-            self.compiled.syntax.clone(),
+            self.compiled.syntax_config.clone(),
+            // TODO: this is not entirely great, but good enough for this use case.
+            Default::default(),
         ) {
             Ok(ast) => find_undeclared(&ast, nested),
             Err(_) => HashSet::new(),
@@ -253,8 +272,8 @@ impl<'env, 'source> Template<'env, 'source> {
     pub fn new_state(&self) -> State<'_, 'env> {
         State::new(
             self.env,
-            Default::default(),
-            self.initial_auto_escape,
+            Context::new(self.env.recursion_limit()),
+            self.compiled.initial_auto_escape,
             &self.compiled.instructions,
             prepare_blocks(&self.compiled.blocks),
         )
@@ -265,7 +284,7 @@ impl<'env, 'source> Template<'env, 'source> {
     ///
     /// For templates loaded as string on the environment this API contract
     /// cannot be upheld because the template might not live long enough.  Under
-    /// normal cirumstances however such a template object would never make it
+    /// normal circumstances however such a template object would never make it
     /// to the callers of this API as this API is used for including or extending,
     /// both of which should only ever get access to a template from the environment
     /// which holds a borrowed ref.
@@ -291,7 +310,7 @@ impl<'env, 'source> Template<'env, 'source> {
     /// Returns the initial auto escape setting.
     #[cfg(feature = "multi_template")]
     pub(crate) fn initial_auto_escape(&self) -> AutoEscape {
-        self.initial_auto_escape
+        self.compiled.initial_auto_escape
     }
 }
 
@@ -321,7 +340,9 @@ pub struct CompiledTemplate<'source> {
     /// Optional size hint for string rendering.
     pub buffer_size_hint: usize,
     /// The syntax config that created it.
-    pub syntax: SyntaxConfig,
+    pub syntax_config: SyntaxConfig,
+    /// The initial setting of auto escaping.
+    pub initial_auto_escape: AutoEscape,
 }
 
 impl<'env> fmt::Debug for CompiledTemplate<'env> {
@@ -341,20 +362,25 @@ impl<'source> CompiledTemplate<'source> {
     pub fn new(
         name: &'source str,
         source: &'source str,
-        syntax: SyntaxConfig,
+        config: &TemplateConfig,
     ) -> Result<CompiledTemplate<'source>, Error> {
-        attach_basic_debug_info(Self::_new_impl(name, source, syntax), source)
+        attach_basic_debug_info(Self::_new_impl(name, source, config), source)
     }
 
     fn _new_impl(
         name: &'source str,
         source: &'source str,
-        syntax: SyntaxConfig,
+        config: &TemplateConfig,
     ) -> Result<CompiledTemplate<'source>, Error> {
         // the parser/compiler combination can create constants in which case
         // we can probably benefit from the value optimization a bit.
         let _guard = value::value_optimization();
-        let ast = ok!(parse_with_syntax(source, name, syntax.clone()));
+        let ast = ok!(parse(
+            source,
+            name,
+            config.syntax_config.clone(),
+            config.ws_config
+        ));
         let mut gen = CodeGenerator::new(name, source);
         gen.compile_stmt(&ast);
         let buffer_size_hint = gen.buffer_size_hint();
@@ -363,7 +389,8 @@ impl<'source> CompiledTemplate<'source> {
             instructions,
             blocks,
             buffer_size_hint,
-            syntax,
+            syntax_config: config.syntax_config.clone(),
+            initial_auto_escape: (config.default_auto_escape)(name),
         })
     }
 }
