@@ -240,10 +240,20 @@ mod builtins {
     use super::*;
 
     use std::borrow::Cow;
-    use std::convert::TryFrom;
 
+    use crate::value::ops::{coerce, CoerceResult};
     use crate::value::ValueKind;
 
+    /// Return true if the object is a boolean value.
+    ///
+    /// ```jinja
+    /// {{ true is boolean }} -> true
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn is_boolean(v: Value) -> bool {
+        v.kind() == ValueKind::Bool
+    }
+
     /// Checks if a value is odd.
     ///
     /// ```jinja
@@ -264,6 +274,20 @@ mod builtins {
         i128::try_from(v).ok().map_or(false, |x| x % 2 == 0)
     }
 
+    /// Return true if the value is divisible by another one.
+    ///
+    /// ```jinja
+    /// {{ 42 is divisibleby(2) }} -> true
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn is_divisibleby(v: Value, other: Value) -> bool {
+        match coerce(&v, &other, false) {
+            Some(CoerceResult::I128(a, b)) => (a % b) == 0,
+            Some(CoerceResult::F64(a, b)) => (a % b) == 0.0,
+            _ => false,
+        }
+    }
+
     /// Checks if this value is a number.
     ///
     /// ```jinja
@@ -275,6 +299,28 @@ mod builtins {
         matches!(v.kind(), ValueKind::Number)
     }
 
+    /// Checks if this value is an integer.
+    ///
+    /// ```jinja
+    /// {{ 42 is integer }} -> true
+    /// {{ 42.0 is integer }} -> false
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn is_integer(v: Value) -> bool {
+        v.is_integer()
+    }
+
+    /// Checks if this value is a float
+    ///
+    /// ```jinja
+    /// {{ 42 is float }} -> false
+    /// {{ 42.0 is float }} -> true
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn is_float(v: Value) -> bool {
+        matches!(v.0, crate::value::ValueRepr::F64(_))
+    }
+
     /// Checks if this value is a string.
     ///
     /// ```jinja
@@ -297,6 +343,17 @@ mod builtins {
         matches!(v.kind(), ValueKind::Seq)
     }
 
+    /// Checks if this value can be iterated over.
+    ///
+    /// ```jinja
+    /// {{ [1, 2, 3] is iterable }} -> true
+    /// {{ 42 is iterable }} -> false
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    pub fn is_iterable(v: Value) -> bool {
+        v.try_iter().is_ok()
+    }
+
     /// Checks if this value is a mapping
     ///
     /// ```jinja
@@ -311,8 +368,8 @@ mod builtins {
     /// Checks if the value is starting with a string.
     ///
     /// ```jinja
-    /// {{ "foobar" is startingwith("foo") }} -> true
-    /// {{ "foobar" is startingwith("bar") }} -> false
+    /// {{ "foobar" is startingwith "foo" }} -> true
+    /// {{ "foobar" is startingwith "bar" }} -> false
     /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn is_startingwith(v: Cow<'_, str>, other: Cow<'_, str>) -> bool {
@@ -322,8 +379,8 @@ mod builtins {
     /// Checks if the value is ending with a string.
     ///
     /// ```jinja
-    /// {{ "foobar" is endingwith("bar") }} -> true
-    /// {{ "foobar" is endingwith("foo") }} -> false
+    /// {{ "foobar" is endingwith "bar" }} -> true
+    /// {{ "foobar" is endingwith "foo" }} -> false
     /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn is_endingwith(v: Cow<'_, str>, other: Cow<'_, str>) -> bool {
@@ -436,10 +493,11 @@ mod builtins {
     /// This is useful when combined with [`select`](crate::filters::select).
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     #[cfg(feature = "builtins")]
-    pub fn is_in(value: &Value, other: &Value) -> bool {
-        crate::value::ops::contains(other, value)
+    pub fn is_in(state: &State, value: &Value, other: &Value) -> Result<bool, Error> {
+        ok!(state.undefined_behavior().assert_iterable(value));
+        Ok(crate::value::ops::contains(other, value)
             .map(|value| value.is_true())
-            .unwrap_or(false)
+            .unwrap_or(false))
     }
 
     /// Checks if a value is `true`.
@@ -485,6 +543,58 @@ mod builtins {
     pub fn is_test(state: &State, name: &str) -> bool {
         state.env.get_test(name).is_some()
     }
+
+    /// Checks if a string is all lowercase.
+    ///
+    /// ```jinja
+    /// {{ 'foo' is lower }} -> true
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    #[cfg(feature = "builtins")]
+    pub fn is_lower(name: &str) -> bool {
+        name.chars().all(|x| x.is_lowercase())
+    }
+
+    /// Checks if a string is all uppercase.
+    ///
+    /// ```jinja
+    /// {{ 'FOO' is upper }} -> true
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    #[cfg(feature = "builtins")]
+    pub fn is_upper(name: &str) -> bool {
+        name.chars().all(|x| x.is_uppercase())
+    }
+
+    /// Checks if two values are identical.
+    ///
+    /// This primarily exists for compatibilith with Jinja2.  It can be seen as a much
+    /// stricter comparison than a regular comparison.  The main difference is that
+    /// values that have the same structure but a different internal object will not
+    /// compare equal.
+    ///
+    /// ```jinja
+    /// {{ [1, 2, 3] is sameas [1, 2, 3] }}
+    ///     -> false
+    ///
+    /// {{ false is sameas false }}
+    ///     -> true
+    /// ```
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    #[cfg(feature = "builtins")]
+    pub fn is_sameas(value: &Value, other: &Value) -> bool {
+        match (value.as_object(), other.as_object()) {
+            (Some(a), Some(b)) => a.is_same_object(b),
+            (None, Some(_)) | (Some(_), None) => false,
+            (None, None) => {
+                if value.kind() != other.kind() || value.is_integer() != other.is_integer() {
+                    false
+                } else {
+                    value == other
+                }
+            }
+        }
+    }
 }
 
 #[cfg(feature = "builtins")]
