@@ -1,22 +1,23 @@
 use std::borrow::Cow;
 use std::cell::RefCell;
-use std::collections::{BTreeMap, HashMap, HashSet};
-use std::convert::TryFrom;
+use std::collections::HashSet;
 use std::ops::{Deref, DerefMut};
+use std::sync::Arc;
 
 use crate::error::{Error, ErrorKind};
 use crate::utils::UndefinedBehavior;
 use crate::value::{
-    Arc, KeyRef, MapType, Object, Packed, SeqObject, StringType, Value, ValueKind, ValueMap,
-    ValueRepr,
+    DynObject, ObjectRepr, Packed, SmallStr, StringType, Value, ValueKind, ValueMap, ValueRepr,
 };
 use crate::vm::State;
 
+use super::{Enumerator, Object};
+
 /// A utility trait that represents the return value of functions and filters.
 ///
 /// It's implemented for the following types:
 ///
-/// * `Rv` where `Rv` implements `Into<Value>`
+/// * `Rv` where `Rv` implements `Into<AnyMapObject>`
 /// * `Result<Rv, Error>` where `Rv` implements `Into<Value>`
 ///
 /// The equivalent for test functions is [`TestResult`](crate::tests::TestResult).
@@ -110,7 +111,10 @@ where
 /// * bytes: [`&[u8]`][`slice`]
 /// * values: [`Value`], `&Value`
 /// * vectors: [`Vec<T>`]
-/// * sequences: [`&dyn SeqObject`](crate::value::SeqObject)
+/// * objects: [`DynObject`], [`Arc<T>`], `&T` (where `T` is an [`Object`])
+/// * serde deserializable: [`ViaDeserialize<T>`](crate::value::deserialize::ViaDeserialize)
+/// * keyword arguments: [`Kwargs`]
+/// * leftover arguments: [`Rest<T>`]
 ///
 /// The type is also implemented for optional values (`Option<T>`) which is used
 /// to encode optional parameters to filters, functions or tests.  Additionally
@@ -223,8 +227,8 @@ macro_rules! tuple_impls {
                 if values.get(idx).is_some() {
                     Err(Error::from(ErrorKind::TooManyArguments))
                 } else {
-                    // SAFETY: this is safe because both no batter what `rest_first` is set to
-                    // either way the variable is set.
+                    // SAFETY: this is safe because both no matter what `rest_first` is set to
+                    // the rest_name variable is set at this point.
                     Ok(($($name,)* unsafe { $rest_name.unwrap_unchecked() },))
                 }
             }
@@ -267,7 +271,16 @@ impl<'a> From<&'a [u8]> for Value {
 impl<'a> From<&'a str> for Value {
     #[inline(always)]
     fn from(val: &'a str) -> Self {
-        ValueRepr::String(Arc::from(val.to_string()), StringType::Normal).into()
+        SmallStr::try_new(val)
+            .map(|small_str| Value(ValueRepr::SmallStr(small_str)))
+            .unwrap_or_else(|| Value::from(val.to_string()))
+    }
+}
+
+impl<'a> From<&'a String> for Value {
+    #[inline(always)]
+    fn from(val: &'a String) -> Self {
+        Value::from(val.as_str())
     }
 }
 
@@ -288,6 +301,12 @@ impl<'a> From<Cow<'a, str>> for Value {
     }
 }
 
+impl From<Arc<str>> for Value {
+    fn from(value: Arc<str>) -> Self {
+        Value(ValueRepr::String(value, StringType::Normal))
+    }
+}
+
 impl From<()> for Value {
     #[inline(always)]
     fn from(_: ()) -> Self {
@@ -297,47 +316,17 @@ impl From<()> for Value {
 
 impl<V: Into<Value>> FromIterator<V> for Value {
     fn from_iter<T: IntoIterator<Item = V>>(iter: T) -> Self {
-        ValueRepr::Seq(Arc::new(iter.into_iter().map(Into::into).collect())).into()
+        Value::from_object(iter.into_iter().map(Into::into).collect::<Vec<Value>>())
     }
 }
 
 impl<K: Into<Value>, V: Into<Value>> FromIterator<(K, V)> for Value {
     fn from_iter<T: IntoIterator<Item = (K, V)>>(iter: T) -> Self {
-        let map = iter
-            .into_iter()
-            .map(|(k, v)| (KeyRef::Value(k.into()), v.into()))
-            .collect();
-        ValueRepr::Map(Arc::new(map), MapType::Normal).into()
-    }
-}
-
-impl<K: Into<Value>, V: Into<Value>> From<BTreeMap<K, V>> for Value {
-    fn from(val: BTreeMap<K, V>) -> Self {
-        val.into_iter().map(|(k, v)| (k.into(), v.into())).collect()
-    }
-}
-
-impl<K: Into<Value>, V: Into<Value>> From<HashMap<K, V>> for Value {
-    fn from(val: HashMap<K, V>) -> Self {
-        val.into_iter().map(|(k, v)| (k.into(), v.into())).collect()
-    }
-}
-
-impl<T: Into<Value>> From<Vec<T>> for Value {
-    fn from(val: Vec<T>) -> Self {
-        val.into_iter().map(|v| v.into()).collect()
-    }
-}
-
-impl<T: Object> From<Arc<T>> for Value {
-    fn from(object: Arc<T>) -> Self {
-        Value::from(object as Arc<dyn Object>)
-    }
-}
-
-impl From<Arc<str>> for Value {
-    fn from(value: Arc<str>) -> Self {
-        Value(ValueRepr::String(value, StringType::Normal))
+        Value::from_object(
+            iter.into_iter()
+                .map(|(k, v)| (k.into(), v.into()))
+                .collect::<ValueMap>(),
+        )
     }
 }
 
@@ -369,7 +358,8 @@ impl From<u128> for Value {
 impl From<char> for Value {
     #[inline(always)]
     fn from(val: char) -> Self {
-        ValueRepr::String(Arc::from(val.to_string()), StringType::Normal).into()
+        let mut buf = [0u8; 4];
+        ValueRepr::SmallStr(SmallStr::try_new(val.encode_utf8(&mut buf)).unwrap()).into()
     }
 }
 
@@ -385,8 +375,7 @@ value_from!(i64, I64);
 value_from!(f32, F64);
 value_from!(f64, F64);
 value_from!(Arc<Vec<u8>>, Bytes);
-value_from!(Arc<Vec<Value>>, Seq);
-value_from!(Arc<dyn Object>, Dynamic);
+value_from!(DynObject, Object);
 
 fn unsupported_conversion(kind: ValueKind, target: &str) -> Error {
     Error::new(
@@ -462,6 +451,12 @@ primitive_try_from!(char, {
             unsupported_conversion(ValueKind::String, "non single character string")
         }))
     },
+    ValueRepr::SmallStr(ref val) => {
+        let mut char_iter = val.as_str().chars();
+        ok!(char_iter.next().filter(|_| char_iter.next().is_none()).ok_or_else(|| {
+            unsupported_conversion(ValueKind::String, "non single character string")
+        }))
+    },
 });
 primitive_try_from!(f32, {
     ValueRepr::U64(val) => val as f32,
@@ -497,6 +492,8 @@ impl TryFrom<Value> for Arc<str> {
     fn try_from(value: Value) -> Result<Self, Self::Error> {
         match value.0 {
             ValueRepr::String(x, _) => Ok(x),
+            ValueRepr::SmallStr(x) => Ok(Arc::from(x.as_str())),
+            ValueRepr::Bytes(ref x) => Ok(Arc::from(String::from_utf8_lossy(x))),
             _ => Err(Error::new(
                 ErrorKind::InvalidOperation,
                 "value is not a string",
@@ -529,26 +526,13 @@ impl<'a> ArgType<'a> for &[u8] {
     }
 }
 
-impl<'a> ArgType<'a> for &dyn SeqObject {
-    type Output = &'a dyn SeqObject;
-
-    fn from_value(value: Option<&'a Value>) -> Result<Self::Output, Error> {
-        match value {
-            Some(value) => value
-                .as_seq()
-                .ok_or_else(|| Error::new(ErrorKind::InvalidOperation, "value is not a sequence")),
-            None => Err(Error::from(ErrorKind::MissingArgument)),
-        }
-    }
-}
-
 impl<'a, T: ArgType<'a>> ArgType<'a> for Option<T> {
     type Output = Option<T::Output>;
 
     fn from_value(value: Option<&'a Value>) -> Result<Self::Output, Error> {
         match value {
             Some(value) => {
-                if value.is_undefined() {
+                if value.is_undefined() || value.is_none() {
                     Ok(None)
                 } else {
                     T::from_value(Some(value)).map(Some)
@@ -575,7 +559,16 @@ impl<'a> ArgType<'a> for Cow<'_, str> {
         match value {
             Some(value) => Ok(match value.0 {
                 ValueRepr::String(ref s, _) => Cow::Borrowed(s as &str),
-                _ => Cow::Owned(value.to_string()),
+                ValueRepr::SmallStr(ref s) => Cow::Borrowed(s.as_str()),
+                _ => {
+                    if value.is_kwargs() {
+                        return Err(Error::new(
+                            ErrorKind::InvalidOperation,
+                            "cannot convert kwargs to string",
+                        ));
+                    }
+                    Cow::Owned(value.to_string())
+                }
             }),
             None => Err(Error::from(ErrorKind::MissingArgument)),
         }
@@ -615,6 +608,34 @@ impl<'a> ArgType<'a> for &[Value] {
     }
 }
 
+impl<'a, T: Object + 'static> ArgType<'a> for &T {
+    type Output = &'a T;
+
+    #[inline(always)]
+    fn from_value(value: Option<&'a Value>) -> Result<Self::Output, Error> {
+        match value {
+            Some(value) => value
+                .downcast_object_ref()
+                .ok_or_else(|| Error::new(ErrorKind::InvalidOperation, "expected object")),
+            None => Err(Error::from(ErrorKind::MissingArgument)),
+        }
+    }
+}
+
+impl<'a, T: Object + 'static> ArgType<'a> for Arc<T> {
+    type Output = Arc<T>;
+
+    #[inline(always)]
+    fn from_value(value: Option<&'a Value>) -> Result<Self::Output, Error> {
+        match value {
+            Some(value) => value
+                .downcast_object()
+                .ok_or_else(|| Error::new(ErrorKind::InvalidOperation, "expected object")),
+            None => Err(Error::from(ErrorKind::MissingArgument)),
+        }
+    }
+}
+
 /// Utility type to capture remaining arguments.
 ///
 /// In some cases you might want to have a variadic function.  In that case
@@ -734,21 +755,46 @@ impl<'a, T: ArgType<'a, Output = T>> ArgType<'a> for Rest<T> {
 /// ```
 #[derive(Debug, Clone)]
 pub struct Kwargs {
-    values: Arc<ValueMap>,
+    pub(crate) values: Arc<KwargsValues>,
     used: RefCell<HashSet<String>>,
 }
 
+#[repr(transparent)]
+#[derive(Default, Debug)]
+pub(crate) struct KwargsValues(ValueMap);
+
+impl Deref for KwargsValues {
+    type Target = ValueMap;
+
+    fn deref(&self) -> &Self::Target {
+        &self.0
+    }
+}
+
+impl KwargsValues {
+    fn as_value_map<'a>(self: &'a Arc<Self>) -> &'a Arc<ValueMap> {
+        // SAFETY: this is safe because of repr(transparent)
+        unsafe { std::mem::transmute(self) }
+    }
+}
+
+impl Object for KwargsValues {
+    fn get_value(self: &Arc<Self>, key: &Value) -> Option<Value> {
+        self.as_value_map().get_value(key)
+    }
+
+    fn enumerate(self: &Arc<Self>) -> Enumerator {
+        self.as_value_map().enumerate()
+    }
+}
+
 impl<'a> ArgType<'a> for Kwargs {
     type Output = Self;
 
     fn from_value(value: Option<&'a Value>) -> Result<Self, Error> {
         match value {
             Some(value) => {
-                if let ValueRepr::Map(ref map, MapType::Kwargs) = value.0 {
-                    Ok(Kwargs::new(map.clone()))
-                } else {
-                    Err(Error::from(ErrorKind::MissingArgument))
-                }
+                Kwargs::extract(value).ok_or_else(|| Error::from(ErrorKind::MissingArgument))
             }
             None => Ok(Kwargs::new(Default::default())),
         }
@@ -759,12 +805,13 @@ impl<'a> ArgType<'a> for Kwargs {
         values: &'a [Value],
         offset: usize,
     ) -> Result<(Self, usize), Error> {
-        if let Some(value) = values.get(offset) {
-            if let ValueRepr::Map(ref map, MapType::Kwargs) = value.0 {
-                return Ok((Kwargs::new(map.clone()), 1));
-            }
-        }
-        Ok((Kwargs::new(Default::default()), 0))
+        let args = values
+            .get(offset)
+            .and_then(Kwargs::extract)
+            .map(|kwargs| (kwargs, 1))
+            .unwrap_or_else(|| (Kwargs::new(Default::default()), 0));
+
+        Ok(args)
     }
 
     fn is_trailing() -> bool {
@@ -773,19 +820,37 @@ impl<'a> ArgType<'a> for Kwargs {
 }
 
 impl Kwargs {
-    fn new(map: Arc<ValueMap>) -> Kwargs {
+    fn new(map: Arc<KwargsValues>) -> Kwargs {
         Kwargs {
             values: map,
             used: RefCell::new(HashSet::new()),
         }
     }
 
+    /// Given a value, extracts the kwargs if there are any.
+    pub(crate) fn extract(value: &Value) -> Option<Kwargs> {
+        value
+            .as_object()
+            .and_then(|x| x.downcast::<KwargsValues>())
+            .map(Kwargs::new)
+    }
+
+    /// Wraps a value map into kwargs.
+    pub(crate) fn wrap(map: ValueMap) -> Value {
+        Value::from_object(KwargsValues(map))
+    }
+
     /// Get a single argument from the kwargs but don't mark it as used.
     pub fn peek<'a, T>(&'a self, key: &'a str) -> Result<T, Error>
     where
         T: ArgType<'a, Output = T>,
     {
-        T::from_value(self.values.get(&KeyRef::Str(key)))
+        T::from_value(self.values.get(&Value::from(key))).map_err(|mut err| {
+            if err.kind() == ErrorKind::MissingArgument && err.detail().is_none() {
+                err.set_detail(format!("missing keyword argument '{}'", key));
+            }
+            err
+        })
     }
 
     /// Gets a single argument from the kwargs and marks it as used.
@@ -820,7 +885,7 @@ impl Kwargs {
 
     /// Checks if a keyword argument exists.
     pub fn has(&self, key: &str) -> bool {
-        self.values.contains_key(&KeyRef::Str(key))
+        self.values.contains_key(&Value::from(key))
     }
 
     /// Iterates over all passed keyword arguments.
@@ -855,11 +920,9 @@ impl FromIterator<(String, Value)> for Kwargs {
     where
         T: IntoIterator<Item = (String, Value)>,
     {
-        Kwargs::new(Arc::new(
-            iter.into_iter()
-                .map(|(k, v)| (KeyRef::Value(Value::from(k)), v))
-                .collect(),
-        ))
+        Kwargs::new(Arc::new(KwargsValues(
+            iter.into_iter().map(|(k, v)| (Value::from(k), v)).collect(),
+        )))
     }
 }
 
@@ -868,17 +931,15 @@ impl<'a> FromIterator<(&'a str, Value)> for Kwargs {
     where
         T: IntoIterator<Item = (&'a str, Value)>,
     {
-        Kwargs::new(Arc::new(
-            iter.into_iter()
-                .map(|(k, v)| (KeyRef::Value(Value::from(k)), v))
-                .collect(),
-        ))
+        Kwargs::new(Arc::new(KwargsValues(
+            iter.into_iter().map(|(k, v)| (Value::from(k), v)).collect(),
+        )))
     }
 }
 
 impl From<Kwargs> for Value {
     fn from(value: Kwargs) -> Self {
-        Value(ValueRepr::Map(value.values, MapType::Kwargs))
+        Value::from_dyn_object(value.values)
     }
 }
 
@@ -888,7 +949,9 @@ impl TryFrom<Value> for Kwargs {
     fn try_from(value: Value) -> Result<Self, Self::Error> {
         match value.0 {
             ValueRepr::Undefined => Ok(Kwargs::new(Default::default())),
-            ValueRepr::Map(ref val, MapType::Kwargs) => Ok(Kwargs::new(val.clone())),
+            ValueRepr::Object(_) => {
+                Kwargs::extract(&value).ok_or_else(|| Error::from(ErrorKind::InvalidOperation))
+            }
             _ => Err(Error::from(ErrorKind::InvalidOperation)),
         }
     }
@@ -921,7 +984,15 @@ impl<'a> ArgType<'a> for String {
 
     fn from_value(value: Option<&'a Value>) -> Result<Self, Error> {
         match value {
-            Some(value) => Ok(value.to_string()),
+            Some(value) => {
+                if value.is_kwargs() {
+                    return Err(Error::new(
+                        ErrorKind::InvalidOperation,
+                        "cannot convert kwargs to string",
+                    ));
+                }
+                Ok(value.to_string())
+            }
             None => Err(Error::from(ErrorKind::MissingArgument)),
         }
     }
@@ -938,11 +1009,13 @@ impl<'a, T: ArgType<'a, Output = T>> ArgType<'a> for Vec<T> {
         match value {
             None => Ok(Vec::new()),
             Some(value) => {
-                let seq = ok!(value
-                    .as_seq()
-                    .ok_or_else(|| { Error::new(ErrorKind::InvalidOperation, "not a sequence") }));
+                let iter = ok!(value
+                    .as_object()
+                    .filter(|x| matches!(x.repr(), ObjectRepr::Seq | ObjectRepr::Iterable))
+                    .and_then(|x| x.try_iter())
+                    .ok_or_else(|| { Error::new(ErrorKind::InvalidOperation, "not iterable") }));
                 let mut rv = Vec::new();
-                for value in seq.iter() {
+                for value in iter {
                     rv.push(ok!(T::from_value_owned(value)));
                 }
                 Ok(rv)
@@ -951,17 +1024,36 @@ impl<'a, T: ArgType<'a, Output = T>> ArgType<'a> for Vec<T> {
     }
 
     fn from_value_owned(value: Value) -> Result<Self, Error> {
-        let seq = ok!(value
-            .as_seq()
-            .ok_or_else(|| { Error::new(ErrorKind::InvalidOperation, "not a sequence") }));
+        let iter = ok!(value
+            .as_object()
+            .filter(|x| matches!(x.repr(), ObjectRepr::Seq | ObjectRepr::Iterable))
+            .and_then(|x| x.try_iter())
+            .ok_or_else(|| { Error::new(ErrorKind::InvalidOperation, "not iterable") }));
         let mut rv = Vec::new();
-        for value in seq.iter() {
+        for value in iter {
             rv.push(ok!(T::from_value_owned(value)));
         }
         Ok(rv)
     }
 }
 
+impl<'a> ArgType<'a> for DynObject {
+    type Output = Self;
+
+    fn from_value(value: Option<&'a Value>) -> Result<Self, Error> {
+        value
+            .ok_or_else(|| Error::from(ErrorKind::MissingArgument))
+            .and_then(|v| Self::from_value_owned(v.clone()))
+    }
+
+    fn from_value_owned(value: Value) -> Result<Self, Error> {
+        value
+            .as_object()
+            .cloned()
+            .ok_or_else(|| Error::new(ErrorKind::InvalidOperation, "not an object"))
+    }
+}
+
 impl From<Value> for String {
     fn from(val: Value) -> Self {
         val.to_string()
@@ -974,6 +1066,21 @@ impl From<usize> for Value {
     }
 }
 
+impl From<isize> for Value {
+    fn from(val: isize) -> Self {
+        Value::from(val as i64)
+    }
+}
+
+impl<I: Into<Value>> From<Option<I>> for Value {
+    fn from(value: Option<I>) -> Self {
+        match value {
+            Some(value) => value.into(),
+            None => Value::from(()),
+        }
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -1003,4 +1110,32 @@ mod tests {
         assert_eq!(kwargs.get::<Value>("foo").unwrap(), Value::from(1));
         assert_eq!(kwargs.get::<Value>("bar").unwrap(), Value::from(2));
     }
+
+    #[test]
+    fn test_kwargs_fails_string_conversion() {
+        let kwargs = Kwargs::from_iter([("foo", Value::from(1)), ("bar", Value::from(2))]);
+        let args = [Value::from(kwargs)];
+
+        let result = from_args::<(String,)>(&args);
+        assert!(result.is_err());
+        assert_eq!(
+            result.unwrap_err().to_string(),
+            "invalid operation: cannot convert kwargs to string"
+        );
+
+        let result = from_args::<(Cow<str>,)>(&args);
+        assert!(result.is_err());
+        assert_eq!(
+            result.unwrap_err().to_string(),
+            "invalid operation: cannot convert kwargs to string"
+        );
+    }
+
+    #[test]
+    fn test_optional_none() {
+        let (one,) = from_args::<(Option<i32>,)>(args!(None::<i32>)).unwrap();
+        assert!(one.is_none());
+        let (one,) = from_args::<(Option<i32>,)>(args!(Some(Value::UNDEFINED))).unwrap();
+        assert!(one.is_none());
+    }
 }
