@@ -5,17 +5,13 @@ use std::sync::Arc;
 
 use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
-use crate::value::{OwnedValueIterator, Value, ValueRepr};
-use crate::vm::closure_object::Closure;
+use crate::value::{Value, ValueIter};
 use crate::vm::loop_object::Loop;
 
-type Locals<'env> = BTreeMap<&'env str, Value>;
+#[cfg(feature = "macros")]
+use crate::vm::closure_object::Closure;
 
-/// The maximum recursion in the VM.  Normally each stack frame
-/// adds one to this counter (eg: every time a frame is added).
-/// However in some situations more depth is pushed if the cost
-/// of the stack frame is higher.
-const MAX_RECURSION: usize = 500;
+type Locals<'env> = BTreeMap<&'env str, Value>;
 
 pub(crate) struct LoopState {
     pub(crate) with_loop_var: bool,
@@ -24,7 +20,7 @@ pub(crate) struct LoopState {
     // first item is the target jump instruction, the second argument
     // tells us if we need to end capturing.
     pub(crate) current_recursion_jump: Option<(usize, bool)>,
-    pub(crate) iterator: OwnedValueIterator,
+    pub(crate) iterator: ValueIter,
     pub(crate) object: Arc<Loop>,
 }
 
@@ -38,6 +34,7 @@ pub(crate) struct Frame<'env> {
     // duplicated into the closure.  Macros declared on that level, then share
     // the closure object to enclose the parent values.  This emulates the
     // behavior of closures in Jinja2.
+    #[cfg(feature = "macros")]
     pub(crate) closure: Option<Arc<Closure>>,
 }
 
@@ -54,17 +51,14 @@ impl<'env> Frame<'env> {
             locals: Locals::new(),
             ctx,
             current_loop: None,
+            #[cfg(feature = "macros")]
             closure: None,
         }
     }
 
-    /// Creates a new frame with the given context and validates the the value is not invalid
+    /// Creates a new frame with the given context and validates the value is not invalid
     pub fn new_checked(root: Value) -> Result<Frame<'env>, Error> {
-        if let ValueRepr::Invalid(ref err) = root.0 {
-            Err(Error::new(ErrorKind::BadSerialization, err.to_string()))
-        } else {
-            Ok(Frame::new(root))
-        }
+        Ok(Frame::new(ok!(root.validate())))
     }
 }
 
@@ -110,7 +104,16 @@ impl Stack {
         self.values.pop().unwrap()
     }
 
-    pub fn slice_top(&mut self, n: usize) -> &[Value] {
+    pub fn reverse_top(&mut self, n: usize) {
+        let start = self.values.len() - n;
+        self.values[start..].reverse();
+    }
+
+    pub fn get_call_args(&mut self, n: Option<u16>) -> &[Value] {
+        let n = match n {
+            Some(n) => n as usize,
+            None => self.pop().as_usize().unwrap(),
+        };
         &self.values[self.values.len() - n..]
     }
 
@@ -137,6 +140,7 @@ impl From<Vec<Value>> for Stack {
 pub(crate) struct Context<'env> {
     stack: Vec<Frame<'env>>,
     outer_stack_depth: usize,
+    recursion_limit: usize,
 }
 
 impl<'env> fmt::Debug for Context<'env> {
@@ -184,19 +188,19 @@ impl<'env> fmt::Debug for Context<'env> {
     }
 }
 
-impl<'env> Default for Context<'env> {
-    fn default() -> Context<'env> {
+impl<'env> Context<'env> {
+    /// Creates an empty context.
+    pub fn new(recursion_limit: usize) -> Context<'env> {
         Context {
             stack: Vec::with_capacity(32),
             outer_stack_depth: 0,
+            recursion_limit,
         }
     }
-}
 
-impl<'env> Context<'env> {
     /// Creates a context
-    pub fn new(frame: Frame<'env>) -> Context<'env> {
-        let mut rv = Context::default();
+    pub fn new_with_frame(frame: Frame<'env>, recursion_limit: usize) -> Context<'env> {
+        let mut rv = Context::new(recursion_limit);
         rv.stack.push(frame);
         rv
     }
@@ -204,8 +208,11 @@ impl<'env> Context<'env> {
     /// Stores a variable in the context.
     pub fn store(&mut self, key: &'env str, value: Value) {
         let top = self.stack.last_mut().unwrap();
-        if let Some(ref closure) = top.closure {
-            closure.store(key, value.clone());
+        #[cfg(feature = "macros")]
+        {
+            if let Some(ref closure) = top.closure {
+                closure.store(key, value.clone());
+            }
         }
         top.locals.insert(key, value);
     }
@@ -217,17 +224,20 @@ impl<'env> Context<'env> {
     /// unfortunate downside is that this has to be done with a `Mutex`.
     #[cfg(feature = "macros")]
     pub fn enclose(&mut self, env: &Environment, key: &str) {
-        self.closure()
+        self.stack
+            .last_mut()
+            .unwrap()
+            .closure
+            .as_mut()
+            .unwrap()
+            .clone()
             .store_if_missing(key, || self.load(env, key).unwrap_or(Value::UNDEFINED));
     }
 
     /// Loads the closure and returns it.
     #[cfg(feature = "macros")]
-    pub fn closure(&mut self) -> Arc<Closure> {
-        let top = self.stack.last_mut().unwrap();
-        top.closure
-            .get_or_insert_with(|| Arc::new(Closure::default()))
-            .clone()
+    pub fn closure(&mut self) -> Option<&Arc<Closure>> {
+        self.stack.last_mut().unwrap().closure.as_ref()
     }
 
     /// Temporarily takes the closure.
@@ -239,17 +249,26 @@ impl<'env> Context<'env> {
     /// This means that if you override a variable referenced by a macro after
     /// including in the parent template, it will not override the value seen by
     /// the macro.
-    #[cfg(feature = "multi_template")]
+    #[cfg(all(feature = "multi_template", feature = "macros"))]
     pub fn take_closure(&mut self) -> Option<Arc<Closure>> {
         self.stack.last_mut().unwrap().closure.take()
     }
 
     /// Puts the closure back.
-    #[cfg(feature = "multi_template")]
+    #[cfg(feature = "macros")]
     pub fn reset_closure(&mut self, closure: Option<Arc<Closure>>) {
         self.stack.last_mut().unwrap().closure = closure;
     }
 
+    /// Return the base context value
+    #[cfg(feature = "macros")]
+    pub fn clone_base(&self) -> Value {
+        self.stack
+            .first()
+            .map(|x| x.ctx.clone())
+            .unwrap_or_default()
+    }
+
     /// Looks up a variable in the context.
     pub fn load(&self, env: &Environment, key: &str) -> Option<Value> {
         for frame in self.stack.iter().rev() {
@@ -261,7 +280,7 @@ impl<'env> Context<'env> {
             // if we are a loop, check if we are looking up the special loop var.
             if let Some(ref l) = frame.current_loop {
                 if l.with_loop_var && key == "loop" {
-                    return Some(Value::from(l.object.clone()));
+                    return Some(Value::from_dyn_object(l.object.clone()));
                 }
             }
 
@@ -330,7 +349,7 @@ impl<'env> Context<'env> {
     }
 
     fn check_depth(&self) -> Result<(), Error> {
-        if self.depth() > MAX_RECURSION {
+        if self.depth() > self.recursion_limit {
             return Err(Error::new(
                 ErrorKind::InvalidOperation,
                 "recursion limit exceeded",
