@@ -5,64 +5,76 @@ use std::sync::Arc;
 use crate::error::{Error, ErrorKind};
 use crate::output::Output;
 use crate::utils::AutoEscape;
-use crate::value::{
-    KeyRef, MapType, Object, ObjectKind, StringType, StructObject, Value, ValueRepr,
-};
+use crate::value::{Enumerator, Kwargs, Object, Value};
 use crate::vm::state::State;
 use crate::vm::Vm;
 
-pub(crate) struct MacroData {
-    pub name: Arc<str>,
-    pub arg_spec: Vec<Arc<str>>,
+pub(crate) struct Macro {
+    pub name: Value,
+    pub arg_spec: Vec<Value>,
     // because values need to be 'static, we can't hold a reference to the
     // instructions that declared the macro.  Instead of that we place the
     // reference to the macro instruction (and the jump offset) in the
     // state under `state.macros`.
     pub macro_ref_id: usize,
+    pub state_id: isize,
     pub closure: Value,
     pub caller_reference: bool,
 }
 
-pub(crate) struct Macro {
-    // the extra level of Arc here is necessary for recursive calls only.
-    // For more information have a look at the call() method.
-    pub data: Arc<MacroData>,
-}
-
 impl fmt::Debug for Macro {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "{self}")
-    }
-}
-
-impl fmt::Display for Macro {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "<macro {}>", self.data.name)
+        write!(f, "<macro {}>", self.name)
     }
 }
 
 impl Object for Macro {
-    fn kind(&self) -> ObjectKind<'_> {
-        ObjectKind::Struct(self)
+    fn enumerate(self: &Arc<Self>) -> Enumerator {
+        Enumerator::Str(&["name", "arguments", "caller"])
     }
 
-    fn call(&self, state: &State, args: &[Value]) -> Result<Value, Error> {
+    fn get_value(self: &Arc<Self>, key: &Value) -> Option<Value> {
+        Some(match some!(key.as_str()) {
+            "name" => self.name.clone(),
+            "arguments" => Value::from_iter(self.arg_spec.iter().cloned()),
+            "caller" => Value::from(self.caller_reference),
+            _ => return None,
+        })
+    }
+
+    fn call(self: &Arc<Self>, state: &State<'_, '_>, args: &[Value]) -> Result<Value, Error> {
+        // we can only call macros that point to loaded template state.
+        if state.id != self.state_id {
+            return Err(Error::new(
+                ErrorKind::InvalidOperation,
+                "cannot call this macro. template state went away.",
+            ));
+        }
+
         let (args, kwargs) = match args.last() {
-            Some(Value(ValueRepr::Map(kwargs, MapType::Kwargs))) => {
-                (&args[..args.len() - 1], Some(kwargs))
-            }
+            Some(last) => match Kwargs::extract(last) {
+                Some(kwargs) => (&args[..args.len() - 1], Some(kwargs)),
+                None => (args, None),
+            },
             _ => (args, None),
         };
 
-        if args.len() > self.data.arg_spec.len() {
+        if args.len() > self.arg_spec.len() {
             return Err(Error::from(ErrorKind::TooManyArguments));
         }
 
         let mut kwargs_used = BTreeSet::new();
-        let mut arg_values = Vec::with_capacity(self.data.arg_spec.len());
-        for (idx, name) in self.data.arg_spec.iter().enumerate() {
-            let kwarg = match kwargs {
-                Some(kwargs) => kwargs.get(&KeyRef::Str(name)),
+        let mut arg_values = Vec::with_capacity(self.arg_spec.len());
+        for (idx, name) in self.arg_spec.iter().enumerate() {
+            let name = match name.as_str() {
+                Some(name) => name,
+                None => {
+                    arg_values.push(Value::UNDEFINED);
+                    continue;
+                }
+            };
+            let kwarg: Option<&Value> = match kwargs {
+                Some(ref kwargs) => kwargs.get(name).ok(),
                 _ => None,
             };
             arg_values.push(match (args.get(idx), kwarg) {
@@ -81,11 +93,12 @@ impl Object for Macro {
             });
         }
 
-        let caller = if self.data.caller_reference {
+        let caller = if self.caller_reference {
             kwargs_used.insert("caller");
             Some(
                 kwargs
-                    .and_then(|x| x.get(&KeyRef::Str("caller")).cloned())
+                    .as_ref()
+                    .and_then(|x| x.get("caller").ok())
                     .unwrap_or(Value::UNDEFINED),
             )
         } else {
@@ -93,7 +106,7 @@ impl Object for Macro {
         };
 
         if let Some(kwargs) = kwargs {
-            for key in kwargs.keys().filter_map(|x| x.as_str()) {
+            for key in kwargs.values.keys().filter_map(|x| x.as_str()) {
                 if !kwargs_used.contains(key) {
                     return Err(Error::new(
                         ErrorKind::TooManyArguments,
@@ -103,17 +116,11 @@ impl Object for Macro {
             }
         }
 
-        let (instructions, offset) = &state.macros[self.data.macro_ref_id];
+        let (instructions, offset) = &state.macros[self.macro_ref_id];
         let vm = Vm::new(state.env());
         let mut rv = String::new();
         let mut out = Output::with_string(&mut rv);
 
-        // If a macro is self referential we need to put a reference to ourselves
-        // there.  Unfortunately because we only have a &self reference here, we
-        // cannot bump our own refcount.  Instead we need to wrap the macro data
-        // into an extra level of Arc to avoid unnecessary clones.
-        let closure = self.data.closure.clone();
-
         // This requires some explanation here.  Because we get the state as &State and
         // not &mut State we are required to create a new state here.  This is unfortunate
         // but makes the calling interface more convenient for the rest of the system.
@@ -123,7 +130,8 @@ impl Object for Macro {
         ok!(vm.eval_macro(
             instructions,
             *offset,
-            closure,
+            self.closure.clone(),
+            state.ctx.clone_base(),
             caller,
             &mut out,
             state,
@@ -136,28 +144,8 @@ impl Object for Macro {
             Value::from(rv)
         })
     }
-}
 
-impl StructObject for Macro {
-    fn static_fields(&self) -> Option<&'static [&'static str]> {
-        Some(&["name", "arguments", "caller"][..])
-    }
-
-    fn get_field(&self, name: &str) -> Option<Value> {
-        match name {
-            "name" => Some(Value(ValueRepr::String(
-                self.data.name.clone(),
-                StringType::Normal,
-            ))),
-            "arguments" => Some(Value::from(
-                self.data
-                    .arg_spec
-                    .iter()
-                    .map(|x| Value(ValueRepr::String(x.clone(), StringType::Normal)))
-                    .collect::<Vec<_>>(),
-            )),
-            "caller" => Some(Value::from(self.data.caller_reference)),
-            _ => None,
-        }
+    fn render(self: &Arc<Self>, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "<macro {}>", self.name)
     }
 }
