@@ -10,18 +10,21 @@ use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
 use crate::output::{CaptureMode, Output};
 use crate::utils::{untrusted_size_hint, AutoEscape, UndefinedBehavior};
+use crate::value::namespace_object::Namespace;
 use crate::value::{
-    ops, value_map_with_capacity, value_optimization, KeyRef, MapType, Value, ValueRepr,
+    ops, value_map_with_capacity, value_optimization, Kwargs, ObjectRepr, Value, ValueMap,
 };
-use crate::vm::context::{Context, Frame, LoopState, Stack};
+use crate::vm::context::{Frame, LoopState, Stack};
 use crate::vm::loop_object::Loop;
 use crate::vm::state::BlockStack;
 
 #[cfg(feature = "macros")]
-use crate::vm::macro_object::{Macro, MacroData};
+use crate::vm::closure_object::Closure;
 
+pub(crate) use crate::vm::context::Context;
 pub use crate::vm::state::State;
 
+#[cfg(feature = "macros")]
 mod closure_object;
 mod context;
 #[cfg(feature = "fuel")]
@@ -37,7 +40,7 @@ const INCLUDE_RECURSION_COST: usize = 10;
 
 // the cost of a single macro call against the stack limit.
 #[cfg(feature = "macros")]
-const MACRO_RECURSION_COST: usize = 5;
+const MACRO_RECURSION_COST: usize = 4;
 
 /// Helps to evaluate something.
 #[cfg_attr(feature = "internal_debug", derive(Debug))]
@@ -91,7 +94,7 @@ impl<'env> Vm<'env> {
         let _guard = value_optimization();
         let mut state = State::new(
             self.env,
-            Context::new(ok!(Frame::new_checked(root))),
+            Context::new_with_frame(ok!(Frame::new_checked(root)), self.env.recursion_limit()),
             auto_escape,
             instructions,
             prepare_blocks(blocks),
@@ -107,17 +110,19 @@ impl<'env> Vm<'env> {
         instructions: &Instructions<'env>,
         pc: usize,
         closure: Value,
+        context_base: Value,
         caller: Option<Value>,
         out: &mut Output,
         state: &State,
         args: Vec<Value>,
     ) -> Result<Option<Value>, Error> {
-        let mut ctx = Context::new(Frame::new(closure));
+        let mut ctx = Context::new_with_frame(Frame::new(context_base), self.env.recursion_limit());
+        ok!(ctx.push_frame(Frame::new(closure)));
         if let Some(caller) = caller {
             ctx.store("caller", caller);
         }
         ok!(ctx.incr_depth(state.ctx.depth() + MACRO_RECURSION_COST));
-        self.eval_impl(
+        self.do_eval(
             &mut State {
                 env: self.env,
                 ctx,
@@ -127,7 +132,11 @@ impl<'env> Vm<'env> {
                 blocks: BTreeMap::default(),
                 loaded_templates: Default::default(),
                 #[cfg(feature = "macros")]
+                id: state.id,
+                #[cfg(feature = "macros")]
                 macros: state.macros.clone(),
+                #[cfg(feature = "macros")]
+                closure_tracker: state.closure_tracker.clone(),
                 #[cfg(feature = "fuel")]
                 fuel_tracker: state.fuel_tracker.clone(),
             },
@@ -144,9 +153,30 @@ impl<'env> Vm<'env> {
         state: &mut State<'_, 'env>,
         out: &mut Output,
     ) -> Result<Option<Value>, Error> {
-        self.eval_impl(state, out, Stack::default(), 0)
+        self.do_eval(state, out, Stack::default(), 0)
     }
 
+    /// Performs the actual evaluation, optionally with stack growth functionality.
+    fn do_eval(
+        &self,
+        state: &mut State<'_, 'env>,
+        out: &mut Output,
+        stack: Stack,
+        pc: usize,
+    ) -> Result<Option<Value>, Error> {
+        #[cfg(feature = "stacker")]
+        {
+            stacker::maybe_grow(32 * 1024, 1024 * 1024, || {
+                self.eval_impl(state, out, stack, pc)
+            })
+        }
+        #[cfg(not(feature = "stacker"))]
+        {
+            self.eval_impl(state, out, stack, pc)
+        }
+    }
+
+    #[inline]
     fn eval_impl(
         &self,
         state: &mut State<'_, 'env>,
@@ -183,6 +213,8 @@ impl<'env> Vm<'env> {
             }};
         }
 
+        // looks nicer this way
+        #[allow(clippy::while_let_loop)]
         loop {
             let instr = match state.instructions.get(pc) {
                 Some(instr) => instr,
@@ -201,14 +233,18 @@ impl<'env> Vm<'env> {
                     };
                     out.end_capture(AutoEscape::None);
                     pc = 0;
+                    // because we swap out the instructions we also need to unload all
+                    // the filters and tests to ensure that we are not accidentally
+                    // reusing the local_ids for completely different filters.
+                    loaded_filters = [None; MAX_LOCALS];
+                    loaded_tests = [None; MAX_LOCALS];
                     continue;
                 }
             };
 
             // if we only have two arguments that we pull from the stack, we
             // can assign them to a and b.  This slightly reduces the amount of
-            // code bloat generated here.  Do the same for a potential error
-            // that needs processing.
+            // code bloat generated here.  Same with the error.
             let a;
             let b;
             let mut err;
@@ -249,10 +285,10 @@ impl<'env> Vm<'env> {
             macro_rules! assert_valid {
                 ($expr:expr) => {{
                     let val = $expr;
-                    if let ValueRepr::Invalid(ref err) = val.0 {
-                        bail!(Error::new(ErrorKind::BadSerialization, err.to_string()));
+                    match val.validate() {
+                        Ok(val) => val,
+                        Err(err) => bail!(err),
                     }
-                    val
                 }};
             }
 
@@ -264,6 +300,12 @@ impl<'env> Vm<'env> {
             }
 
             match instr {
+                Instruction::Swap => {
+                    let a = stack.pop();
+                    let b = stack.pop();
+                    stack.push(a);
+                    stack.push(b);
+                }
                 Instruction::EmitRaw(val) => {
                     // this only produces a format error, no need to attach
                     // location information.
@@ -292,6 +334,18 @@ impl<'env> Vm<'env> {
                         None => ctx_ok!(undefined_behavior.handle_undefined(a.is_undefined())),
                     });
                 }
+                Instruction::SetAttr(name) => {
+                    b = stack.pop();
+                    a = stack.pop();
+                    if let Some(ns) = b.downcast_object_ref::<Namespace>() {
+                        ns.set_value(name, a);
+                    } else {
+                        bail!(Error::new(
+                            ErrorKind::InvalidOperation,
+                            format!("can only assign to namespaces, not {}", b.kind())
+                        ));
+                    }
+                }
                 Instruction::GetItem => {
                     a = stack.pop();
                     b = stack.pop();
@@ -315,45 +369,78 @@ impl<'env> Vm<'env> {
                 }
                 Instruction::BuildMap(pair_count) => {
                     let mut map = value_map_with_capacity(*pair_count);
+                    stack.reverse_top(*pair_count * 2);
                     for _ in 0..*pair_count {
-                        let value = stack.pop();
                         let key = stack.pop();
-                        map.insert(KeyRef::Value(key), value);
+                        let value = stack.pop();
+                        map.insert(key, value);
                     }
-                    stack.push(Value(ValueRepr::Map(map.into(), MapType::Normal)))
+                    stack.push(Value::from_object(map))
                 }
                 Instruction::BuildKwargs(pair_count) => {
                     let mut map = value_map_with_capacity(*pair_count);
+                    stack.reverse_top(*pair_count * 2);
                     for _ in 0..*pair_count {
-                        let value = stack.pop();
                         let key = stack.pop();
-                        map.insert(KeyRef::Value(key), value);
+                        let value = stack.pop();
+                        map.insert(key, value);
                     }
-                    stack.push(Value(ValueRepr::Map(map.into(), MapType::Kwargs)))
+                    stack.push(Kwargs::wrap(map))
                 }
-                Instruction::BuildList(count) => {
-                    let mut v = Vec::with_capacity(untrusted_size_hint(*count));
+                Instruction::MergeKwargs(count) => {
+                    let mut kwargs_sources = Vec::new();
                     for _ in 0..*count {
+                        kwargs_sources.push(stack.pop());
+                    }
+                    kwargs_sources.reverse();
+                    let values: &[Value] = &kwargs_sources;
+                    let mut rv = ValueMap::new();
+                    for value in values {
+                        ctx_ok!(self.env.undefined_behavior().assert_iterable(value));
+                        let iter = ctx_ok!(value
+                            .as_object()
+                            .filter(|x| x.repr() == ObjectRepr::Map)
+                            .and_then(|x| x.try_iter_pairs())
+                            .ok_or_else(|| {
+                                Error::new(
+                                    ErrorKind::InvalidOperation,
+                                    format!(
+                                        "attempted to apply keyword arguments from non map (got {})",
+                                        value.kind()
+                                    ),
+                                )
+                            }));
+                        for (key, value) in iter {
+                            rv.insert(key, value);
+                        }
+                    }
+                    stack.push(Kwargs::wrap(rv));
+                }
+                Instruction::BuildList(n) => {
+                    let count = n.unwrap_or_else(|| stack.pop().try_into().unwrap());
+                    let mut v = Vec::with_capacity(untrusted_size_hint(count));
+                    for _ in 0..count {
                         v.push(stack.pop());
                     }
                     v.reverse();
-                    stack.push(Value(ValueRepr::Seq(Arc::new(v))));
+                    stack.push(Value::from_object(v))
                 }
                 Instruction::UnpackList(count) => {
-                    ctx_ok!(self.unpack_list(&mut stack, count));
+                    ctx_ok!(self.unpack_list(&mut stack, *count));
                 }
-                Instruction::ListAppend => {
-                    a = stack.pop();
-                    // this intentionally only works with actual sequences
-                    if let ValueRepr::Seq(mut v) = stack.pop().0 {
-                        Arc::make_mut(&mut v).push(a);
-                        stack.push(Value(ValueRepr::Seq(v)))
-                    } else {
-                        bail!(Error::new(
-                            ErrorKind::InvalidOperation,
-                            "cannot append to non-list"
-                        ));
+                Instruction::UnpackLists(count) => {
+                    let mut lists = Vec::new();
+                    for _ in 0..*count {
+                        lists.push(stack.pop());
                     }
+                    let mut len = 0;
+                    for list in lists.into_iter().rev() {
+                        for item in ctx_ok!(list.try_iter()) {
+                            stack.push(item);
+                            len += 1;
+                        }
+                    }
+                    stack.push(Value::from(len));
                 }
                 Instruction::Add => func_binop!(add),
                 Instruction::Sub => func_binop!(sub),
@@ -449,13 +536,13 @@ impl<'env> Vm<'env> {
                 }
                 Instruction::JumpIfFalse(jump_target) => {
                     a = stack.pop();
-                    if !a.is_true() {
+                    if !ctx_ok!(undefined_behavior.is_true(&a)) {
                         pc = *jump_target;
                         continue;
                     }
                 }
                 Instruction::JumpIfFalseOrPop(jump_target) => {
-                    if !stack.peek().is_true() {
+                    if !ctx_ok!(undefined_behavior.is_true(stack.peek())) {
                         pc = *jump_target;
                         continue;
                     } else {
@@ -463,7 +550,7 @@ impl<'env> Vm<'env> {
                     }
                 }
                 Instruction::JumpIfTrueOrPop(jump_target) => {
-                    if stack.peek().is_true() {
+                    if ctx_ok!(undefined_behavior.is_true(stack.peek())) {
                         pc = *jump_target;
                         continue;
                     } else {
@@ -501,9 +588,10 @@ impl<'env> Vm<'env> {
                                 format!("filter {name} is unknown"),
                             )
                         }));
-                    let args = stack.slice_top(*arg_count);
+                    let args = stack.get_call_args(*arg_count);
+                    let arg_count = args.len();
                     a = ctx_ok!(filter.apply_to(state, args));
-                    stack.drop_top(*arg_count);
+                    stack.drop_top(arg_count);
                     stack.push(a);
                 }
                 Instruction::PerformTest(name, arg_count, local_id) => {
@@ -513,53 +601,58 @@ impl<'env> Vm<'env> {
                     .ok_or_else(|| {
                         Error::new(ErrorKind::UnknownTest, format!("test {name} is unknown"))
                     }));
-                    let args = stack.slice_top(*arg_count);
+                    let args = stack.get_call_args(*arg_count);
+                    let arg_count = args.len();
                     let rv = ctx_ok!(test.perform(state, args));
-                    stack.drop_top(*arg_count);
+                    stack.drop_top(arg_count);
                     stack.push(Value::from(rv));
                 }
                 Instruction::CallFunction(name, arg_count) => {
+                    let args = stack.get_call_args(*arg_count);
                     // super is a special function reserved for super-ing into blocks.
-                    if *name == "super" {
-                        if *arg_count != 0 {
+                    let rv = if *name == "super" {
+                        if !args.is_empty() {
                             bail!(Error::new(
                                 ErrorKind::InvalidOperation,
                                 "super() takes no arguments",
                             ));
                         }
-                        stack.push(ctx_ok!(self.perform_super(state, out, true)));
+                        ctx_ok!(self.perform_super(state, out, true))
                     // loop is a special name which when called recurses the current loop.
                     } else if *name == "loop" {
-                        if *arg_count != 1 {
+                        if args.len() != 1 {
                             bail!(Error::new(
                                 ErrorKind::InvalidOperation,
-                                format!("loop() takes one argument, got {}", *arg_count)
+                                "loop() takes one argument"
                             ));
                         }
-                        // leave the one argument on the stack for the recursion
+                        // leave the one argument on the stack for the recursion.  The
+                        // recurse_loop! macro itself will perform a jump and not return here.
                         recurse_loop!(true);
                     } else if let Some(func) = state.lookup(name) {
-                        let args = stack.slice_top(*arg_count);
-                        a = ctx_ok!(func.call(state, args));
-                        stack.drop_top(*arg_count);
-                        stack.push(a);
+                        ctx_ok!(func.call(state, args))
                     } else {
                         bail!(Error::new(
                             ErrorKind::UnknownFunction,
                             format!("{name} is unknown"),
                         ));
-                    }
+                    };
+                    let arg_count = args.len();
+                    stack.drop_top(arg_count);
+                    stack.push(rv);
                 }
                 Instruction::CallMethod(name, arg_count) => {
-                    let args = stack.slice_top(*arg_count);
+                    let args = stack.get_call_args(*arg_count);
+                    let arg_count = args.len();
                     a = ctx_ok!(args[0].call_method(state, name, &args[1..]));
-                    stack.drop_top(*arg_count);
+                    stack.drop_top(arg_count);
                     stack.push(a);
                 }
                 Instruction::CallObject(arg_count) => {
-                    let args = stack.slice_top(*arg_count);
+                    let args = stack.get_call_args(*arg_count);
+                    let arg_count = args.len();
                     a = ctx_ok!(args[0].call(state, &args[1..]));
-                    stack.drop_top(*arg_count);
+                    stack.drop_top(arg_count);
                     stack.push(a);
                 }
                 Instruction::DupTop => {
@@ -619,9 +712,9 @@ impl<'env> Vm<'env> {
                     let locals = state.ctx.current_locals_mut();
                     let mut module = value_map_with_capacity(locals.len());
                     for (key, value) in locals.iter() {
-                        module.insert(KeyRef::Value(Value::from(*key)), value.clone());
+                        module.insert(Value::from(*key), value.clone());
                     }
-                    stack.push(Value(ValueRepr::Map(module.into(), MapType::Normal)));
+                    stack.push(Value::from_object(module));
                 }
                 #[cfg(feature = "macros")]
                 Instruction::BuildMacro(name, offset, flags) => {
@@ -631,11 +724,24 @@ impl<'env> Vm<'env> {
                 Instruction::Return => break,
                 #[cfg(feature = "macros")]
                 Instruction::Enclose(name) => {
+                    // the first time we enclose a value, we need to create a closure
+                    // and store it on the context, and add it to the closure tracker
+                    // for cycle breaking.
+                    if state.ctx.closure().is_none() {
+                        let closure = Arc::new(Closure::default());
+                        state.closure_tracker.track_closure(closure.clone());
+                        state.ctx.reset_closure(Some(closure));
+                    }
                     state.ctx.enclose(state.env, name);
                 }
                 #[cfg(feature = "macros")]
                 Instruction::GetClosure => {
-                    stack.push(Value::from(state.ctx.closure()));
+                    stack.push(
+                        state
+                            .ctx
+                            .closure()
+                            .map_or(Value::UNDEFINED, |x| Value::from_dyn_object(x.clone())),
+                    );
                 }
             }
             pc += 1;
@@ -652,22 +758,23 @@ impl<'env> Vm<'env> {
         out: &mut Output,
         ignore_missing: bool,
     ) -> Result<(), Error> {
-        use crate::value::SeqObject;
-
-        let single_name_slice = std::slice::from_ref(&name);
-        let choices = name
-            .as_seq()
-            .unwrap_or(&single_name_slice as &dyn SeqObject);
+        let obj = name.as_object();
+        let choices = obj
+            .as_ref()
+            .and_then(|d| d.try_iter())
+            .into_iter()
+            .flatten()
+            .chain(obj.is_none().then(|| name.clone()));
 
         let mut templates_tried = vec![];
-        for choice in choices.iter() {
+        for choice in choices {
             let name = ok!(choice.as_str().ok_or_else(|| {
                 Error::new(
                     ErrorKind::InvalidOperation,
                     "template name was not a string",
                 )
             }));
-            let tmpl = match self.env.get_template(name) {
+            let tmpl = match state.get_template(name) {
                 Ok(tmpl) => tmpl,
                 Err(err) => {
                     if err.kind() == ErrorKind::TemplateNotFound {
@@ -683,11 +790,24 @@ impl<'env> Vm<'env> {
             let old_escape = mem::replace(&mut state.auto_escape, tmpl.initial_auto_escape());
             let old_instructions = mem::replace(&mut state.instructions, new_instructions);
             let old_blocks = mem::replace(&mut state.blocks, prepare_blocks(new_blocks));
-            let old_closure = state.ctx.take_closure();
+            // we need to make a copy of the loaded templates here as we want
+            // to forget about the templates that an include triggered by the
+            // time the include finishes.
+            let old_loaded_templates = state.loaded_templates.clone();
             ok!(state.ctx.incr_depth(INCLUDE_RECURSION_COST));
-            let rv = self.eval_state(state, out);
-            state.ctx.reset_closure(old_closure);
+            let rv;
+            #[cfg(feature = "macros")]
+            {
+                let old_closure = state.ctx.take_closure();
+                rv = self.eval_state(state, out);
+                state.ctx.reset_closure(old_closure);
+            }
+            #[cfg(not(feature = "macros"))]
+            {
+                rv = self.eval_state(state, out);
+            }
             state.ctx.decr_depth(INCLUDE_RECURSION_COST);
+            state.loaded_templates = old_loaded_templates;
             state.auto_escape = old_escape;
             state.instructions = old_instructions;
             state.blocks = old_blocks;
@@ -798,14 +918,14 @@ impl<'env> Vm<'env> {
                 format!("cycle in template inheritance. {name:?} was referenced more than once"),
             ));
         }
-        let tmpl = ok!(self.env.get_template(name));
+        let tmpl = ok!(state.get_template(name));
         let (new_instructions, new_blocks) = ok!(tmpl.instructions_and_blocks());
         state.loaded_templates.insert(new_instructions.name());
         for (name, instr) in new_blocks.iter() {
             state
                 .blocks
                 .entry(name)
-                .or_insert_with(BlockStack::default)
+                .or_default()
                 .append_instructions(instr);
         }
         Ok(new_instructions)
@@ -868,7 +988,14 @@ impl<'env> Vm<'env> {
     ) -> Result<(), Error> {
         #[allow(unused_mut)]
         let mut iterator = ok!(state.undefined_behavior().try_iter(iterable));
-        let len = iterator.len();
+        // for an iterator where the lower and upper bound are matching we can
+        // consider them to have ExactSizeIterator semantics.  We do however not
+        // expect ExactSizeIterator bounds themselves to support iteration by
+        // other means.
+        let len = match iterator.size_hint() {
+            (lower, Some(upper)) if lower == upper => Some(lower),
+            _ => None,
+        };
         let depth = state
             .ctx
             .current_loop()
@@ -896,25 +1023,28 @@ impl<'env> Vm<'env> {
         Ok(())
     }
 
-    fn unpack_list(&self, stack: &mut Stack, count: &usize) -> Result<(), Error> {
+    fn unpack_list(&self, stack: &mut Stack, count: usize) -> Result<(), Error> {
         let top = stack.pop();
-        let seq = ok!(top
-            .as_seq()
-            .ok_or_else(|| Error::new(ErrorKind::CannotUnpack, "not a sequence")));
-        if seq.item_count() != *count {
-            return Err(Error::new(
-                ErrorKind::CannotUnpack,
-                format!(
-                    "sequence of wrong length (expected {}, got {})",
-                    *count,
-                    seq.item_count()
-                ),
-            ));
-        }
-        for item in seq.iter().rev() {
+        let iter = ok!(top
+            .as_object()
+            .and_then(|x| x.try_iter())
+            .ok_or_else(|| Error::new(ErrorKind::CannotUnpack, "value is not iterable")));
+
+        let mut n = 0;
+        for item in iter {
             stack.push(item);
+            n += 1;
+        }
+
+        if n == count {
+            stack.reverse_top(n);
+            Ok(())
+        } else {
+            Err(Error::new(
+                ErrorKind::CannotUnpack,
+                format!("sequence of wrong length (expected {}, got {})", count, n,),
+            ))
         }
-        Ok(())
     }
 
     #[cfg(feature = "macros")]
@@ -926,29 +1056,19 @@ impl<'env> Vm<'env> {
         name: &str,
         flags: u8,
     ) {
-        use crate::compiler::instructions::MACRO_CALLER;
+        use crate::{compiler::instructions::MACRO_CALLER, vm::macro_object::Macro};
 
-        let arg_spec = match stack.pop().0 {
-            ValueRepr::Seq(args) => args
-                .iter()
-                .map(|value| match &value.0 {
-                    ValueRepr::String(arg, _) => arg.clone(),
-                    _ => unreachable!(),
-                })
-                .collect(),
-            _ => unreachable!(),
-        };
+        let arg_spec = stack.pop().try_iter().unwrap().collect();
         let closure = stack.pop();
         let macro_ref_id = state.macros.len();
         Arc::make_mut(&mut state.macros).push((state.instructions, offset));
         stack.push(Value::from_object(Macro {
-            data: Arc::new(MacroData {
-                name: Arc::from(name.to_string()),
-                arg_spec,
-                macro_ref_id,
-                closure,
-                caller_reference: (flags & MACRO_CALLER) != 0,
-            }),
+            name: Value::from(name),
+            arg_spec,
+            macro_ref_id,
+            state_id: state.id,
+            closure,
+            caller_reference: (flags & MACRO_CALLER) != 0,
         }));
     }
 }
