@@ -5,6 +5,7 @@ use crate::compiler::instructions::Instructions;
 use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
 use crate::output::Output;
+use crate::template::Template;
 use crate::utils::{AutoEscape, UndefinedBehavior};
 use crate::value::{ArgType, Value};
 use crate::vm::context::Context;
@@ -12,6 +13,14 @@ use crate::vm::context::Context;
 #[cfg(feature = "fuel")]
 use crate::vm::fuel::FuelTracker;
 
+/// When macros are used, the state carries an `id` counter.  Whenever a state is
+/// created, the counter is incremented.  This exists because macros can keep a reference
+/// to instructions from another state by index.  Without this counter it would
+/// be possible for a macro to be called with a different state (different id)
+/// which mean we likely panic.
+#[cfg(feature = "macros")]
+static STATE_ID: std::sync::atomic::AtomicIsize = std::sync::atomic::AtomicIsize::new(0);
+
 /// Provides access to the current execution state of the engine.
 ///
 /// A read only reference is passed to filter functions and similar objects to
@@ -39,7 +48,11 @@ pub struct State<'template, 'env> {
     #[allow(unused)]
     pub(crate) loaded_templates: BTreeSet<&'env str>,
     #[cfg(feature = "macros")]
+    pub(crate) id: isize,
+    #[cfg(feature = "macros")]
     pub(crate) macros: std::sync::Arc<Vec<(&'template Instructions<'env>, usize)>>,
+    #[cfg(feature = "macros")]
+    pub(crate) closure_tracker: std::sync::Arc<crate::vm::closure_object::ClosureTracker>,
     #[cfg(feature = "fuel")]
     pub(crate) fuel_tracker: Option<std::sync::Arc<FuelTracker>>,
 }
@@ -66,6 +79,8 @@ impl<'template, 'env> State<'template, 'env> {
         blocks: BTreeMap<&'env str, BlockStack<'template, 'env>>,
     ) -> State<'template, 'env> {
         State {
+            #[cfg(feature = "macros")]
+            id: STATE_ID.fetch_add(1, std::sync::atomic::Ordering::Relaxed),
             env,
             ctx,
             current_block: None,
@@ -75,6 +90,8 @@ impl<'template, 'env> State<'template, 'env> {
             loaded_templates: BTreeSet::new(),
             #[cfg(feature = "macros")]
             macros: Default::default(),
+            #[cfg(feature = "macros")]
+            closure_tracker: Default::default(),
             #[cfg(feature = "fuel")]
             fuel_tracker: env.fuel().map(FuelTracker::new),
         }
@@ -84,7 +101,7 @@ impl<'template, 'env> State<'template, 'env> {
     pub(crate) fn new_for_env(env: &'env Environment) -> State<'env, 'env> {
         State::new(
             env,
-            Context::default(),
+            Context::new(env.recursion_limit()),
             AutoEscape::None,
             &crate::compiler::instructions::EMPTY_INSTRUCTIONS,
             BTreeMap::new(),
@@ -121,6 +138,13 @@ impl<'template, 'env> State<'template, 'env> {
     }
 
     /// Looks up a variable by name in the context.
+    ///
+    /// # Note on Closures
+    ///
+    /// Macros and call blocks analyze which variables are referenced and
+    /// create closures for them.  This means that unless a variable is defined
+    /// as a [global](Environment::add_global) in the environment or it was
+    /// referenced by a macro, this method won't be able to find it.
     #[inline(always)]
     pub fn lookup(&self, name: &str) -> Option<Value> {
         self.ctx.load(self.env, name)
@@ -196,11 +220,24 @@ impl<'template, 'env> State<'template, 'env> {
         self.ctx.exports().keys().copied().collect()
     }
 
+    /// Fetches a template by name with path joining.
+    ///
+    /// This works like [`Environment::get_template`] with the difference that the lookup
+    /// undergoes path joining.  If the environment has a configured path joining callback,
+    /// it will be invoked with the name of the current template as parent template.
+    ///
+    /// For more information see [`Environment::set_path_join_callback`].
+    pub fn get_template(&self, name: &str) -> Result<Template<'env, 'env>, Error> {
+        self.env
+            .get_template(&self.env.join_template_path(name, self.name()))
+    }
+
     /// Invokes a filter with some arguments.
     ///
     /// ```
     /// # use minijinja::Environment;
     /// # let mut env = Environment::new();
+    /// # env.add_filter("upper", |x: &str| x.to_uppercase());
     /// # let tmpl = env.template_from_str("").unwrap();
     /// # let state = tmpl.new_state();
     /// let rv = state.apply_filter("upper", &["hello world".into()]).unwrap();
@@ -218,6 +255,7 @@ impl<'template, 'env> State<'template, 'env> {
     /// ```
     /// # use minijinja::Environment;
     /// # let mut env = Environment::new();
+    /// # env.add_test("even", |x: i32| x % 2 == 0);
     /// # let tmpl = env.template_from_str("").unwrap();
     /// # let state = tmpl.new_state();
     /// let rv = state.perform_test("even", &[42i32.into()]).unwrap();
